from collections import OrderedDict
from contextlib import ExitStack
from torch import optim
import torch
import time
import os

from .beta_crown import print_optimized_beta
from .cuda_utils import double2float
from .utils import logger, reduction_sum

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .bound_general import BoundedModule

default_optimize_bound_args = {
    'enable_alpha_crown': True,  # Enable optimization of alpha.
    'enable_beta_crown': False,  # Enable beta split constraint.
    'iteration': 20,  # Number of alpha/beta optimization iterations.
    # Share some alpha variables to save memory at the cost of slightly
    # looser bounds.
    'use_shared_alpha': False,
    # Optimizer used for alpha and beta optimization.
    'optimizer': 'adam',
    # Save best results of alpha/beta/bounds during optimization.
    'keep_best': True,
    # Only optimize bounds of last layer during alpha/beta CROWN.
    'fix_interm_bounds': True,
    # Learning rate for the optimizable parameter alpha in alpha-CROWN.
    'lr_alpha': 0.5,
    # Learning rate for the optimizable parameter beta in beta-CROWN.
    'lr_beta': 0.05,
    # Initial alpha variables by calling CROWN once.
    'init_alpha': True,
    'lr_coeffs': 0.01,  # Learning rate for coeffs for refinement
    # Layers to be refined, separated by commas.
    # -1 means preactivation before last activation.
    'intermediate_refinement_layers': [-1],
    # When batch size is not 1, this reduction function is applied to
    # reduce the bounds into a scalar.
    'loss_reduction_func': reduction_sum,
    # Criteria function of early stop.
    'stop_criterion_func': lambda x: False,
    # Learning rate decay factor during bounds optimization.
    'lr_decay': 0.98,
    # Number of iterations that we will start considering early stop
    # if tracking no improvement.
    'early_stop_patience': 10,
    # Start to save optimized best bounds
    # when current_iteration > int(iteration*start_save_best)
    'start_save_best': 0.5,
    # Use double fp (float64) at the last iteration in alpha/beta CROWN.
    'use_float64_in_last_iteration': False,
    # Use the newly fixed loss function. By default, it is set to False
    # for compatibility with existing use cases.
    # Try to ensure that the parameters always match with the optimized bounds.
    'deterministic': False,
}


def opt_reuse(self: 'BoundedModule'):
    for node in self.get_enabled_opt_act():
        node.opt_reuse()


def opt_no_reuse(self: 'BoundedModule'):
    for node in self.get_enabled_opt_act():
        node.opt_no_reuse()


def _set_alpha(optimizable_activations, parameters, alphas, lr):
    """Set best_alphas, alphas and parameters list."""
    for node in optimizable_activations:
        alphas.extend(list(node.alpha.values()))
        node.opt_start()
    # Alpha has shape (2, output_shape, batch_dim, node_shape)
    parameters.append({'params': alphas, 'lr': lr, 'batch_dim': 2})
    # best_alpha is a dictionary of dictionary. Each key is the alpha variable
    # for one activation layer, and each value is a dictionary contains all
    # activation layers after that layer as keys.
    best_alphas = OrderedDict()
    for m in optimizable_activations:
        best_alphas[m.name] = {}
        for alpha_m in m.alpha:
            best_alphas[m.name][alpha_m] = m.alpha[alpha_m].detach().clone()
            # We will directly replace the dictionary for each activation layer after
            # optimization, so the saved alpha might not have require_grad=True.
            m.alpha[alpha_m].requires_grad_()

    return best_alphas


def _set_gammas(nodes, parameters):
    """
    Adds gammas to parameters list
    """
    gammas = []
    for node in nodes:
        if hasattr(node, 'gammas'):
            gammas.append(node.gammas)
    gamma_lr = 0.1
    parameters.append({'params': gammas, 'lr': gamma_lr})

def _save_ret_first_time(bounds, fill_value, x, best_ret):
    """Save results at the first iteration to best_ret."""
    if bounds is not None:
        best_bounds = torch.full_like(
            bounds, fill_value=fill_value, device=x[0].device, dtype=x[0].dtype)
    else:
        best_bounds = None

    if bounds is not None:
        best_ret.append(bounds.detach().clone())
    else:
        best_ret.append(None)

    return best_bounds


def _to_float64(self: 'BoundedModule', C, x, aux_reference_bounds, interm_bounds):
    """
    Transfer variables to float64 only in the last iteration to help alleviate
    floating point error.
    """
    self.to(torch.float64)
    C = C.to(torch.float64)
    x = self._to(x, torch.float64)
    # best_intermediate_bounds is linked to aux_reference_bounds!
    # we only need call .to() for one of them
    self._to(aux_reference_bounds, torch.float64, inplace=True)
    interm_bounds = self._to(
        interm_bounds, torch.float64)

    return C, x, interm_bounds


def _to_default_dtype(self: 'BoundedModule', x, total_loss, full_ret, ret,
                      best_intermediate_bounds, return_A):
    """
    Switch back to default precision from float64 typically to adapt to
    afterwards operations.
    """
    total_loss = total_loss.to(torch.get_default_dtype())
    self.to(torch.get_default_dtype())
    x[0].to(torch.get_default_dtype())
    full_ret = list(full_ret)
    if isinstance(ret[0], torch.Tensor):
        # round down lower bound
        full_ret[0] = double2float(full_ret[0], 'down')
    if isinstance(ret[1], torch.Tensor):
        # round up upper bound
        full_ret[1] = double2float(full_ret[1], 'up')
    for _k, _v in best_intermediate_bounds.items():
        _v[0] = double2float(_v[0], 'down')
        _v[1] = double2float(_v[1], 'up')
        best_intermediate_bounds[_k] = _v
    if return_A:
        full_ret[2] = self._to(full_ret[2], torch.get_default_dtype())

    return total_loss, x, full_ret


def _get_idx_mask(idx, full_ret_bound, best_ret_bound, loss_reduction_func):
    """Get index for improved elements."""
    assert idx in [0, 1], (
        '0 means updating lower bound, 1 means updating upper bound')
    if idx == 0:
        idx_mask = (loss_reduction_func(full_ret_bound)
                    > loss_reduction_func(best_ret_bound)).view(-1)
    else:
        idx_mask = (loss_reduction_func(full_ret_bound)
                    < loss_reduction_func(best_ret_bound)).view(-1)
    improved_idx = None
    if idx_mask.any():
        # we only pick up the results improved in a batch
        improved_idx = idx_mask.nonzero(as_tuple=True)[0]
    return idx_mask, improved_idx


def _update_best_ret(full_ret_bound, best_ret_bound, full_ret, best_ret,
                     need_update, loss_reduction_func, idx, deterministic=False):
    """Update best_ret_bound and best_ret by comparing with new results."""
    assert idx in [0, 1], (
        '0 means updating lower bound, 1 means updating upper bound')
    idx_mask, improved_idx = _get_idx_mask(
        idx, full_ret_bound, best_ret_bound, loss_reduction_func)

    if improved_idx is not None:
        need_update = True
        compare = torch.max if idx == 0 else torch.min
        if not deterministic:
            best_ret_bound[improved_idx] = compare(
                full_ret_bound[improved_idx], best_ret_bound[improved_idx])
        else:
            best_ret_bound[improved_idx] = full_ret_bound[improved_idx]
        if full_ret[idx] is not None:
            if not deterministic:
                best_ret[idx][improved_idx] = compare(
                    full_ret[idx][improved_idx],
                    best_ret[idx][improved_idx])
            else:
                best_ret[idx][improved_idx] = full_ret[idx][improved_idx]

    return best_ret_bound, best_ret, need_update, idx_mask, improved_idx


def _update_optimizable_activations(
        optimizable_activations, interm_bounds,
        fix_interm_bounds, best_intermediate_bounds,
        reference_idx, idx, alpha, best_alphas, deterministic):
    """
    Update bounds and alpha of optimizable_activations.
    """
    for node in optimizable_activations:
        # Update best intermediate layer bounds only when they are optimized.
        # If they are already fixed in interm_bounds, then do
        # nothing.
        if (interm_bounds is None
                or node.inputs[0].name not in interm_bounds
                or not fix_interm_bounds):
            if deterministic:
                best_intermediate_bounds[node.name][0][idx] = node.inputs[0].lower[reference_idx]
                best_intermediate_bounds[node.name][1][idx] = node.inputs[0].upper[reference_idx]
            else:
                best_intermediate_bounds[node.name][0][idx] = torch.max(
                    best_intermediate_bounds[node.name][0][idx],
                    node.inputs[0].lower[reference_idx])
                best_intermediate_bounds[node.name][1][idx] = torch.min(
                    best_intermediate_bounds[node.name][1][idx],
                    node.inputs[0].upper[reference_idx])
        if alpha:
            # Each alpha has shape (2, output_shape, batch, *shape) for act.
            # For other activation function this can be different.
            for alpha_m in node.alpha:
                best_alphas[node.name][alpha_m][:, :,
                    idx] = node.alpha[alpha_m][:, :, idx]


def update_best_beta(self: 'BoundedModule', enable_opt_interm_bounds, betas,
                     best_betas, idx):
    """
    Update best beta by given idx.
    """
    if enable_opt_interm_bounds and betas:
        for node in self.splittable_activations:
            for node_input in node.inputs:
                for key in node_input.sparse_betas.keys():
                    best_betas[node_input.name][key] = (
                        node_input.sparse_betas[key].val.detach().clone())
    else:
        for node in self.nodes_with_beta:
            best_betas[node.name][idx] = node.sparse_betas[0].val[idx]


def _get_optimized_bounds(
        self: 'BoundedModule', x=None, aux=None, C=None, IBP=False,
        forward=False, method='backward', bound_side='lower',
        reuse_ibp=False, return_A=False, average_A=False, final_node_name=None,
        interm_bounds=None, reference_bounds=None,
        aux_reference_bounds=None, needed_A_dict=None,
        decision_thresh=None):
    """
    Optimize CROWN lower/upper bounds by alpha and/or beta.
    """

    opts = self.bound_opts['optimize_bound_args']
    iteration = opts['iteration']
    beta = opts['enable_beta_crown']
    alpha = opts['enable_alpha_crown']
    opt_choice = opts['optimizer']
    keep_best = opts['keep_best']
    fix_interm_bounds = opts['fix_interm_bounds']
    loss_reduction_func = opts['loss_reduction_func']
    stop_criterion_func = opts['stop_criterion_func']
    use_float64_in_last_iteration = opts['use_float64_in_last_iteration']
    early_stop_patience = opts['early_stop_patience']
    start_save_best = opts['start_save_best']
    deterministic = opts['deterministic']
    enable_opt_interm_bounds = self.bound_opts.get(
        'enable_opt_interm_bounds', False)
    sparse_intermediate_bounds = self.bound_opts.get(
        'sparse_intermediate_bounds', False)
    verbosity = self.bound_opts['verbosity']

    if bound_side not in ['lower', 'upper']:
        raise ValueError(bound_side)
    bound_lower = bound_side == 'lower'
    bound_upper = bound_side == 'upper'

    assert alpha or beta, (
        'nothing to optimize, use compute bound instead!')

    if C is not None:
        self.final_shape = C.size()[:2]
        self.bound_opts.update({'final_shape': self.final_shape})
    if opts['init_alpha']:
        # TODO: this should set up aux_reference_bounds.
        self.init_alpha(x, share_alphas=opts['use_shared_alpha'],
                        method=method, c=C, final_node_name=final_node_name)

    optimizable_activations = self.get_enabled_opt_act()

    alphas, parameters = [], []
    dense_coeffs_mask = []
    if alpha:
        best_alphas = _set_alpha(
            optimizable_activations, parameters, alphas, opts['lr_alpha'])
    if beta:
        ret_set_beta = self.set_beta(
            enable_opt_interm_bounds=enable_opt_interm_bounds, parameters=parameters,
            lr_beta=opts['lr_beta'], dense_coeffs_mask=dense_coeffs_mask)
        betas, best_betas, coeffs, dense_coeffs_mask = ret_set_beta[:4]

    start = time.time()

    if isinstance(decision_thresh, torch.Tensor):
        if decision_thresh.dim() == 1:
            # add the spec dim to be aligned with compute_bounds return
            decision_thresh = decision_thresh.unsqueeze(-1)

    if opt_choice == 'adam':
        opt = optim.Adam(parameters)
    elif opt_choice == 'sgd':
        opt = optim.SGD(parameters, momentum=0.9)
    else:
        raise NotImplementedError(opt_choice)

    # Create a weight vector to scale learning rate.
    loss_weight = torch.ones(size=(x[0].size(0),), device=x[0].device)
    scheduler = optim.lr_scheduler.ExponentialLR(opt, opts['lr_decay'])

    # best_intermediate_bounds is linked to aux_reference_bounds!
    best_intermediate_bounds = {}
    if (sparse_intermediate_bounds and aux_reference_bounds is None
            and reference_bounds is not None):
        aux_reference_bounds = {}
        for name, (lb, ub) in reference_bounds.items():
            aux_reference_bounds[name] = [
                lb.detach().clone(), ub.detach().clone()]
    if aux_reference_bounds is None:
        aux_reference_bounds = {}

    need_grad = True
    patience = 0
    for i in range(iteration):

        intermediate_constr = None

        if not fix_interm_bounds:
            # If we still optimize all intermediate neurons, we can use
            # interm_bounds as reference bounds.
            if reference_bounds is None:
                reference_bounds = {}
            if interm_bounds is not None:
                reference_bounds.update(interm_bounds)
            interm_bounds = {}

        if i == iteration - 1:
            # No grad update needed for the last iteration
            need_grad = False
            if (self.device == 'cuda'
                    and torch.get_default_dtype() == torch.float32
                    and use_float64_in_last_iteration):
                C, x, interm_bounds = self._to_float64(
                    C, x, aux_reference_bounds, interm_bounds)

        with torch.no_grad() if not need_grad else ExitStack():
            # ret is lb, ub or lb, ub, A_dict (if return_A is set to true)
            ret = self.compute_bounds(
                x, aux, C, method=method, IBP=IBP, forward=forward,
                bound_lower=bound_lower, bound_upper=bound_upper,
                reuse_ibp=reuse_ibp, return_A=return_A,
                final_node_name=final_node_name, average_A=average_A,
                # When intermediate bounds are recomputed, we must set it
                # to None
                interm_bounds=interm_bounds if fix_interm_bounds else None,
                # This is the currently tightest interval, which will be used to
                # pass split constraints when intermediate betas are used.
                reference_bounds=reference_bounds,
                # This is the interval used for checking for unstable neurons.
                aux_reference_bounds=aux_reference_bounds if sparse_intermediate_bounds else None,
                # These are intermediate layer beta variables and their
                # corresponding A matrices and biases.
                intermediate_constr=intermediate_constr,
                needed_A_dict=needed_A_dict,
                update_mask=None)
        ret_l, ret_u = ret[0], ret[1]

        if i == 0:
            # save results at the first iteration
            best_ret = []
            best_ret_l = _save_ret_first_time(
                ret[0], float('-inf'), x, best_ret)
            best_ret_u = _save_ret_first_time(
                ret[1], float('inf'), x, best_ret)
            ret_0 = ret[0].detach().clone() if bound_lower else ret[1].detach().clone()

            for node in optimizable_activations:
                new_intermediate = [node.inputs[0].lower.detach().clone(),
                                    node.inputs[0].upper.detach().clone()]
                best_intermediate_bounds[node.name] = new_intermediate
                if sparse_intermediate_bounds:
                    # Always using the best bounds so far as the reference
                    # bounds.
                    aux_reference_bounds[node.inputs[0].name] = new_intermediate

        l = ret_l
        # Reduction over the spec dimension.
        if ret_l is not None and ret_l.shape[1] != 1:
            l = loss_reduction_func(ret_l)
        u = ret_u
        if ret_u is not None and ret_u.shape[1] != 1:
            u = loss_reduction_func(ret_u)

        # full_l, full_ret_l and full_u, full_ret_u is used for update the best
        full_ret_l, full_ret_u = ret_l, ret_u
        full_l = l
        full_ret = ret

        stop_criterion = (stop_criterion_func(full_ret_l) if bound_lower else stop_criterion_func(-full_ret_u))

        loss_ = l if bound_lower else -u
        total_loss = -1 * loss_

        if type(stop_criterion) == bool:
            loss = total_loss.sum() * (not stop_criterion)
        else:
            assert total_loss.shape == stop_criterion.shape
            loss = (total_loss * stop_criterion.logical_not()).sum()
        # For logging, print the total sum. Otherwise the loss may appear
        # to be increasing as more examples are stopped.
        loss_sum = total_loss.sum()

        stop_criterion_final = isinstance(
            stop_criterion, torch.Tensor) and stop_criterion.all()

        if i == iteration - 1:
            best_ret = list(best_ret)
            if best_ret[0] is not None:
                best_ret[0] = best_ret[0].to(torch.get_default_dtype())
            if best_ret[1] is not None:
                best_ret[1] = best_ret[1].to(torch.get_default_dtype())

        if (i == iteration - 1 and self.device == 'cuda'
                and torch.get_default_dtype() == torch.float32
                and use_float64_in_last_iteration):
            total_loss, x, full_ret = self._to_default_dtype(
                x, total_loss, full_ret, ret, best_intermediate_bounds, return_A)

        with torch.no_grad():
            # for lb and ub, we update them in every iteration since updating
            # them is cheap
            need_update = False
            if keep_best:
                if best_ret_u is not None:
                    best_ret_u, best_ret, need_update, idx_mask, improved_idx = _update_best_ret(
                        full_ret_u, best_ret_u, full_ret, best_ret, need_update,
                        loss_reduction_func, idx=1, deterministic=deterministic)
                if best_ret_l is not None:
                    best_ret_l, best_ret, need_update, idx_mask, improved_idx = _update_best_ret(
                        full_ret_l, best_ret_l, full_ret, best_ret, need_update,
                        loss_reduction_func, idx=0, deterministic=deterministic)
            else:
                # Not saving the best, just keep the last iteration.
                if full_ret[0] is not None:
                    best_ret[0] = full_ret[0]
                if full_ret[1] is not None:
                    best_ret[1] = full_ret[1]
            if return_A:
                # FIXME: A should also be updated by idx.
                best_ret = [best_ret[0], best_ret[1], full_ret[2]]

            if need_update:
                patience = 0  # bounds improved, reset patience
            else:
                patience += 1

            # Save variables if this is the best iteration.
            # To save computational cost, we only check keep_best at the first
            # (in case divergence) and second half iterations
            # or before early stop by either stop_criterion or
            # early_stop_patience reached
            if (i < 1 or i > int(iteration * start_save_best) or deterministic
                    or stop_criterion_final or patience == early_stop_patience):

                # compare with the first iteration results and get improved indexes
                if bound_lower:
                    if deterministic:
                        idx = improved_idx
                    else:
                        idx_mask, idx = _get_idx_mask(
                            0, full_ret_l, ret_0, loss_reduction_func)
                    ret_0[idx] = full_ret_l[idx]
                else:
                    if deterministic:
                        idx = improved_idx
                    else:
                        idx_mask, idx = _get_idx_mask(
                            1, full_ret_u, ret_0, loss_reduction_func)
                    ret_0[idx] = full_ret_u[idx]

                if idx is not None:
                    # for update propose, we condition the idx to update only
                    # on domains preserved
                    reference_idx = idx

                    _update_optimizable_activations(
                        optimizable_activations, interm_bounds,
                        fix_interm_bounds, best_intermediate_bounds,
                        reference_idx, idx, alpha, best_alphas, deterministic)

                    if beta:
                        self.update_best_beta(enable_opt_interm_bounds, betas,
                                              best_betas, idx)


        if os.environ.get('AUTOLIRPA_DEBUG_OPT', False):
            print(f'****** iter [{i}]',
                  f'loss: {loss_sum.item()}, lr: {opt.param_groups[0]["lr"]}',
            )

        if stop_criterion_final:
            # print(f'\nall verified at {i}th iter')
            break

        if patience > early_stop_patience:
            logger.debug(
                f'Early stop at {i}th iter due to {early_stop_patience}'
                ' iterations no improvement!')
            break

        if i != iteration - 1 and not loss.requires_grad:
            assert i == 0, (i, iteration)
            print('[WARNING] No optimizable parameters found. Will skip optimiziation. '
                  'This happens e.g. if all optimizable layers are freezed or the '
                  'network has no optimizable layers.')
            break

        opt.zero_grad(set_to_none=True)

        if verbosity > 2:
            current_lr = [param_group['lr'] for param_group in opt.param_groups]
            print(f'*** iter [{i}]\n', f'loss: {loss.item()}',
                  total_loss.squeeze().detach().cpu().numpy(), 'lr: ',
                  current_lr)
            if beta:
                print_optimized_beta(optimizable_activations)
            if beta and i == 0 and verbosity > 2:
                breakpoint()

        if i != iteration - 1:
            # we do not need to update parameters in the last step since the
            # best result already obtained
            loss.backward()

            # All intermediate variables are not needed at this point.
            self._clear_and_set_new(None)
            if opt_choice == 'adam-autolr':
                opt.step(lr_scale=[loss_weight, loss_weight])
            else:
                opt.step()

            scheduler.step()

        if beta:
            for b in betas:
                b.data = (b >= 0) * b.data
            for dmi in range(len(dense_coeffs_mask)):
                # apply dense mask to the dense split coeffs matrix
                coeffs[dmi].data = (
                    dense_coeffs_mask[dmi].float() * coeffs[dmi].data)

        if alpha:
            for m in optimizable_activations:
                m.clip_alpha()

    if verbosity > 3:
        breakpoint()

    if keep_best:
        # Set all variables to their saved best values.
        with torch.no_grad():
            for idx, node in enumerate(optimizable_activations):
                if alpha:
                    # Assigns a new dictionary.
                    node.alpha = best_alphas[node.name]
                # Update best intermediate layer bounds only when they are
                # optimized. If they are already fixed in
                # interm_bounds, then do nothing.
                best_intermediate = best_intermediate_bounds[node.name]
                node.inputs[0].lower.data = best_intermediate[0].data
                node.inputs[0].upper.data = best_intermediate[1].data
            if beta:
                for node in self.nodes_with_beta:
                    assert getattr(node, 'sparse_betas', None) is not None
                    if enable_opt_interm_bounds:
                        for key in node.sparse_betas.keys():
                            node.sparse_betas[key].val.copy_(
                                best_betas[node.name][key])
                    else:
                        node.sparse_betas[0].val.copy_(best_betas[node.name])

    if interm_bounds is not None and not fix_interm_bounds:
        for l in self._modules.values():
            if (l.name in interm_bounds.keys()
                    and hasattr(l, 'lower')):
                l.lower = torch.max(l.lower, interm_bounds[l.name][0])
                l.upper = torch.min(l.upper, interm_bounds[l.name][1])
                infeasible_neurons = l.lower > l.upper
                if infeasible_neurons.any():
                    print(f'Infeasibility detected in layer {l.name}.',
                          infeasible_neurons.sum().item(),
                          infeasible_neurons.nonzero()[:, 0])

    if verbosity > 0:
        if best_ret_l is not None:
            # FIXME: unify the handling of l and u.
            print('best_l after optimization:', best_ret_l.sum().item())
            if beta:
                print('beta sum per layer:', [p.sum().item() for p in betas])
        print('alpha/beta optimization time:', time.time() - start)

    for node in optimizable_activations:
        node.opt_end()

    if os.environ.get('AUTOLIRPA_DEBUG_OPT', False):
        print()

    return best_ret


def init_alpha(self: 'BoundedModule', x, share_alphas=False, method='backward',
               c=None, bound_lower=True, bound_upper=True, final_node_name=None,
               interm_bounds=None, activation_opt_params=None,
               skip_bound_compute=False):
    self(*x) # Do a forward pass to set perturbed nodes
    final = (self.final_node() if final_node_name is None
             else self[final_node_name])
    self._set_used_nodes(final)

    optimizable_activations = self.get_enabled_opt_act()
    for node in optimizable_activations:
        # TODO(7/6/2023) In the future, we may need to enable alpha sharing
        # automatically by consider the size of all the optimizable nodes in the
        # graph. For now, only an adhoc check in MatMul is added.
        node._all_optimizable_activations = optimizable_activations

        # initialize the parameters
        node.opt_init()

    if (not skip_bound_compute or interm_bounds is None or
            activation_opt_params is None or not all(
                [act.name in activation_opt_params
                 for act in self.optimizable_activations])):
        skipped = False
        # if new interval is None, then CROWN interval is not present
        # in this case, we still need to redo a CROWN pass to initialize
        # lower/upper
        with torch.no_grad():
            # We temporarilly deactivate output constraints
            l, u = self.compute_bounds(
                x=x, C=c, method=method, bound_lower=bound_lower,
                bound_upper=bound_upper, final_node_name=final_node_name,
                interm_bounds=interm_bounds)
    else:
        # we skip, but we still would like to figure out the "used",
        # "perturbed", "backward_from" of each note in the graph
        skipped = True
        # this set the "perturbed" property
        self.set_input(*x, interm_bounds=interm_bounds)
        self.backward_from = {node: [final] for node in self._modules}

    final_node_name = final_node_name or self.final_name

    init_intermediate_bounds = {}
    for node in optimizable_activations:
        start_nodes = []
        if method in ['forward', 'forward+backward']:
            start_nodes.append(('_forward', 1, None, False))
        if method in ['backward', 'forward+backward']:
            backward_from_node = node
            start_nodes += self.get_alpha_crown_start_nodes(
                node,
                c=c,
                share_alphas=share_alphas,
                final_node_name=final_node_name,
                backward_from_node=backward_from_node
            )
        if skipped:
            node.restore_optimized_params(activation_opt_params[node.name])
        else:
            node.init_opt_parameters(start_nodes)
        if node in self.splittable_activations:
            for i in node.requires_input_bounds:
                input_node = node.inputs[i]
                if not input_node.perturbed:
                    continue
                init_intermediate_bounds[node.inputs[i].name] = (
                    [node.inputs[i].lower.detach(),
                    node.inputs[i].upper.detach()])

    if self.bound_opts['verbosity'] >= 1:
        print('Optimizable variables initialized.')
    if skip_bound_compute:
        return init_intermediate_bounds
    else:
        return l, u, init_intermediate_bounds



def get_refined_interm_bounds(self: 'BoundedModule'):
    interm_bounds = {
        node.name: [
            node.lower.to(self.device) if hasattr(node, 'lower') else None,
            node.upper.to(self.device) if hasattr(node, 'upper') else None,
        ] for node in self.split_nodes
    }

    if any([(None in _ ) for _ in interm_bounds.values()]):
        return None

    return interm_bounds



from __future__ import annotations
from beartype import beartype
import typing
import torch
import copy

from ..abstractor.utils import _append_tensor
from ..solver.sat_solver import SATSolver
from ..util.misc.logger import logger
from ..util.misc.result import AbstractResults
from ..auto_LiRPA.bound_ops import *

if typing.TYPE_CHECKING:
    from .. import abstractor
    from .. import heuristic


# @beartype
def compute_masks(lower_bounds: dict, upper_bounds: dict, device: str, non_blocking: bool = True) -> dict:
    # TODO: shifted relu should not use 0.0, e.g., max(x, 1.0)
    new_masks = {
        j: torch.logical_and(
                    lower_bounds[j] < 0,
                    upper_bounds[j] > 0).flatten(start_dim=1).to(torch.get_default_dtype()).to(device=device, non_blocking=non_blocking)
        for j in lower_bounds
    }
    return new_masks


# @beartype
def _compute_babsr_scores(abstractor: 'abstractor.abstractor.NetworkAbstractor',
                          lower_bounds: dict, upper_bounds: dict,
                          lAs: dict, masks: dict,
                          reduce_op: typing.Callable,
                          batch: int, number_bounds: int) -> tuple[list, list]:
    score = []
    intercept_tb = []

    # last to first layer
    for layer in reversed(abstractor.net.split_nodes):
        assert len(abstractor.net.split_activations[layer.name]) == 1
        # layer data
        this_layer_mask = masks[layer.name].unsqueeze(1)
        pre_act_layer = abstractor.net.split_activations[layer.name][0][0]
        assert len(pre_act_layer.inputs) == 1
        ratio = lAs[pre_act_layer.name]

        # ratio
        ratio_temp_0, ratio_temp_1 = _compute_ratio(lower_bounds[layer.name], upper_bounds[layer.name])

        # intercept scores, backup scores, lower score is better
        intercept_temp = torch.clamp(ratio, max=0)
        intercept_candidate = intercept_temp * ratio_temp_1.unsqueeze(1)

        # (batch, neuron)
        reshaped_intercept_candidate = intercept_candidate.view(batch, number_bounds, -1) * this_layer_mask
        intercept_tb.insert(0, reshaped_intercept_candidate.mean(1))

        # bias
        b_temp = _get_bias_term(pre_act_layer.inputs[0], ratio)

        # branching scores, higher score is better
        ratio_temp_0 = ratio_temp_0.unsqueeze(1)
        bias_candidate_1 = b_temp * (ratio_temp_0 - 1)
        bias_candidate_2 = b_temp * ratio_temp_0
        bias_candidate = reduce_op(bias_candidate_1, bias_candidate_2)
        score_candidate = bias_candidate + intercept_candidate

        # (batch, neuron)
        score_candidate = score_candidate.abs().view(batch, number_bounds, -1) * this_layer_mask
        score.insert(0, score_candidate.mean(1))
    return score, intercept_tb


# @beartype
def _history_to_clause(h: dict, name_mapping: dict) -> list:
    clause = []
    for lname, ldata in h.items():
        assert sum(ldata[2]) == 0 # ReLU only
        # extract data
        var_names, signs = ldata[0], ldata[1]
        # convert data
        var_names = [name_mapping[lname, int(v)] for v in var_names]
        # negation
        signs = [-int(s) for s in signs]
        # append literals
        clause += [v * s for v, s in zip(var_names, signs)]
    return clause

# @beartype
def _compute_ratio(lower_bound: torch.Tensor, upper_bound: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    lower_temp = lower_bound.clamp(max=0)
    upper_temp = upper_bound.clamp(min=0)
    slope_ratio = upper_temp / (upper_temp - lower_temp)
    intercept = -1 * lower_temp * slope_ratio
    return slope_ratio, intercept


# @beartype
def _get_bias_term(input_node, ratio: torch.Tensor) -> torch.Tensor:
    if type(input_node) == BoundConv:
        if len(input_node.inputs) > 2:
            bias = input_node.inputs[-1].param.detach().unsqueeze(-1).unsqueeze(-1)
        else:
            bias = 0
    elif type(input_node) == BoundLinear:
        # TODO: consider if no bias
        bias = input_node.inputs[-1].param.detach()
    elif type(input_node) == BoundAdd:
        bias = 0
        for l in input_node.inputs:
            if type(l) == BoundConv:
                if len(l.inputs) > 2:
                    bias += l.inputs[-1].param.detach().unsqueeze(-1).unsqueeze(-1)
            if type(l) == BoundBatchNormalization:
                bias += 0
            if type(l) == BoundAdd:
                for ll in l.inputs:
                    if type(ll) == BoundConv:
                        bias += ll.inputs[-1].param.detach().unsqueeze(-1).unsqueeze(-1)
    elif type(input_node) == BoundBatchNormalization:
        bias = input_node.inputs[-3].param.detach().view(-1, *([1] * (ratio.ndim - 3)))
    else:
        print(type(input_node))
        raise NotImplementedError()

    return bias * ratio


# @beartype
def update_hidden_bounds_histories(self: 'heuristic.domains_list.DomainsList', lower_bounds: dict, upper_bounds: dict,
                                   histories: list, literal: int, batch_idx: int | torch.Tensor) -> bool:
    assert literal != 0
    # extract decision
    lid, nid = self.reversed_var_mapping[abs(literal)]

    # update histories
    history = histories[batch_idx]
    loc = _append_tensor(history[lid][0], nid, dtype=torch.long)
    sign = _append_tensor(history[lid][1], +1 if literal > 0 else -1)
    beta = _append_tensor(history[lid][2], 0.0) # ReLU only
    history[lid] = (loc, sign, beta)


    # update bounds
    if literal > 0: # active neuron
        lower_bounds[lid][batch_idx].flatten()[nid] = 0.0 # ReLU only
    else: # inactive neuron
        upper_bounds[lid][batch_idx].flatten()[nid] = 0.0 # ReLU only

    if upper_bounds[lid][batch_idx].flatten()[nid] < lower_bounds[lid][batch_idx].flatten()[nid]:
        return False
    return True


# @beartype
def create_one(masks: dict, clauses: list, var_mapping: dict) -> tuple[SATSolver | None, list]:
    # literals
    literals_to_assign = []
    for lname, lmask in masks.items():
        for nid, nstatus in enumerate(lmask.flatten()):
            if nstatus != 0:
                literal = int(nstatus * var_mapping[lname, nid])
                literals_to_assign.append(literal)

    # create sat solver
    new_sat_solver = SATSolver(clauses)

    # assign
    if not new_sat_solver.multiple_assign(literals_to_assign):
        return None, [] # conflict

    # propagation
    bcp_stat, bcp_vars = new_sat_solver.bcp()
    if not bcp_stat:
        return None, [] # conflict

    return new_sat_solver, bcp_vars


# @beartype
def init_sat_solver(self: 'heuristic.domains_list.DomainsList', objective_ids: torch.Tensor, remain_idx: torch.Tensor,
                    lower_bounds: dict, upper_bounds: dict, histories: list, preconditions: dict) -> torch.Tensor:
    assert torch.equal(objective_ids, torch.unique(objective_ids))
    # initial learned conflict clauses
    clauses_per_objective = {k: [_history_to_clause(c, self.var_mapping) for c in v] for k, v in preconditions.items()}
    # pprint(clauses_per_objective)

    # masks: 1 for active, -1 for inactive, 0 for unstable
    masks = {
        k: ((lower_bounds[k] > 0).flatten(1).int() - (upper_bounds[k] < 0).flatten(1).int()).detach().cpu()
            for k in lower_bounds if k != self.net.final_name
    }

    masks_per_objective = {int(objective_id): {k: v[batch_id] for k, v in masks.items()} for (batch_id, objective_id) in enumerate(objective_ids)}
    # print(masks_per_objective)

    self.all_sat_solvers = []
    new_remain_idx = []
    for (batch_id, objective_id) in enumerate(objective_ids):
        if batch_id not in remain_idx:
            continue

        new_sat_solver, bcp_vars = create_one(
            masks=masks_per_objective[int(objective_id)],
            clauses=clauses_per_objective[int(objective_id)],
            var_mapping=self.var_mapping,
        )

        update_stats = [self.update_hidden_bounds_histories(
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            histories=histories,
            literal=literal,
            batch_idx=batch_id,
        ) for literal in bcp_vars] + [True]

        # print(batch_id, bcp_vars)
        if (new_sat_solver is not None) and all(update_stats):
            # save
            self.all_sat_solvers.append(new_sat_solver)
            new_remain_idx.append(batch_id)

    return torch.tensor(new_remain_idx)


# @beartype
def boolean_propagation(self: 'heuristic.domains_list.DomainsList', domain_params: AbstractResults, decisions: list, batch_idx: int | torch.Tensor) -> SATSolver | None:
    assert len(decisions) * 2 == len(domain_params.input_lowers) == len(domain_params.sat_solvers)

    # new solver
    new_sat_solver = copy.deepcopy(domain_params.sat_solvers[batch_idx])

    # new decision
    n_name, n_id, n_point = decisions[batch_idx % len(decisions)]
    assert n_point == 0.0 # relu only
    variable = self.var_mapping[n_name, n_id]
    literal = variable if batch_idx < len(decisions) else -variable

    # assign
    if not new_sat_solver.assign(literal):
        logger.debug('[!] Assign conflicted')
        return None

    # propagation
    bcp_stat, bcp_vars = new_sat_solver.bcp()
    if not bcp_stat: # conflict
        return None

    if len(bcp_vars):
        # print('BCP', bcp_vars, 'batch_idx', batch_idx)
        update_stats = [self.update_hidden_bounds_histories(
                            lower_bounds=domain_params.lower_bounds,
                            upper_bounds=domain_params.upper_bounds,
                            histories=domain_params.histories,
                            literal=lit,
                            batch_idx=batch_idx)
                        for lit in bcp_vars]

        if not all(update_stats):
            logger.debug('[!] BCP assign conflicted')
            return None

    return new_sat_solver


# @beartype
def save_conflict_clauses(self: 'heuristic.domains_list.DomainsList', domain_params: AbstractResults, remaining_index: torch.Tensor) -> None:
    assert domain_params.objective_ids is not None
    for i in range(len(domain_params.histories)):
        if i in remaining_index:
            continue
        self.all_conflict_clauses[int(domain_params.objective_ids[i])].append(domain_params.histories[i])



import os
import sys
import torch
from torch.utils.cpp_extension import load, BuildExtension, CUDAExtension
from setuptools import setup

class DummyCudaClass:
    """A dummy class with error message when a CUDA function is called."""
    def __getattr__(self, attr):
        if attr == "double2float":
            # When CUDA module is not built successfully, use a workaround.
            def _f(x, d):
                print('WARNING: Missing CUDA kernels. Please enable CUDA build by setting environment variable AUTOLIRPA_ENABLE_CUDA_BUILD=1 for the correct behavior!')
                return x.float()
            return _f
        def _f(*args, **kwargs):
            raise RuntimeError(f"method {attr} not available because CUDA module was not built.")
        return _f

if __name__ == "__main__" and len(sys.argv) > 1:
    # Build and install native CUDA modules that can be directly imported later
    print('Building and installing native CUDA modules...')
    setup(
        name='auto_LiRPA_cuda_utils',
        ext_modules=[CUDAExtension('auto_LiRPA_cuda_utils', [
            'auto_LiRPA/cuda/cuda_utils.cpp',
            'auto_LiRPA/cuda/cuda_kernels.cu'
        ])],
        cmdclass={'build_ext': BuildExtension.with_options()},
    )
    exit(0)

if torch.cuda.is_available() and os.environ.get('AUTOLIRPA_ENABLE_CUDA_BUILD', False):
    try:
        import auto_LiRPA_cuda_utils as _cuda_utils
    except:
        print('CUDA modules have not been installed')
        try:
            print('Building native CUDA modules...')
            code_dir = os.path.dirname(os.path.abspath(__file__))
            verbose = os.environ.get('AUTOLIRPA_DEBUG_CUDA_BUILD', None) is not None
            _cuda_utils = load(
                'cuda_utils', [os.path.join(code_dir, 'cuda', 'cuda_utils.cpp'), os.path.join(code_dir, 'cuda', 'cuda_kernels.cu')], verbose=verbose)
            print('CUDA modules have been built.')
        except:
            print('CUDA module build failure. Some features will be unavailable.')
            print('Please make sure the latest CUDA toolkit is installed in your system.')
            if verbose:
                print(sys.exc_info()[2])
            else:
                print('Set environment variable AUTOLIRPA_DEBUG_CUDA_BUILD=1 to view build log.')
            _cuda_utils = DummyCudaClass()
else:
    if os.environ.get('AUTOLIRPA_ENABLE_CUDA_BUILD', False):
        print('CUDA unavailable. Some features are disabled.')
    _cuda_utils = DummyCudaClass()

double2float = _cuda_utils.double2float

def test_double2float():
    # Test the double2float function.
    import time
    shape = (3,4,5)

    a = torch.randn(size=shape, dtype=torch.float64, device='cuda')
    a = a.transpose(0,1)

    au = _cuda_utils.double2float(a, "up")
    ad = _cuda_utils.double2float(a, "down")

    print(a.size(), au.size(), ad.size())

    a_flatten = a.reshape(-1)
    au_flatten = au.reshape(-1)
    ad_flatten = ad.reshape(-1)

    for i in range(a_flatten.numel()):
        ai = a_flatten[i].item()
        aui = au_flatten[i].item()
        adi = ad_flatten[i].item()
        print(adi, ai, aui)
        assert adi <= ai
        assert aui >= ai
    del a, au, ad, a_flatten, au_flatten, ad_flatten

    # Performance benchmark.
    for j in [1, 4, 16, 64, 256, 1024]:
        shape = (j, 512, 1024)
        print(f'shape: {shape}')
        t = torch.randn(size=shape, dtype=torch.float64, device='cuda')

        torch.cuda.synchronize()
        start_time = time.time()
        for i in range(10):
            tt = t.float()
        torch.cuda.synchronize()
        del tt
        pytorch_time = time.time() - start_time
        print(f'pytorch rounding time: {pytorch_time:.4f}')

        torch.cuda.synchronize()
        start_time = time.time()
        for i in range(10):
            tu = _cuda_utils.double2float(t, "up")
        torch.cuda.synchronize()
        del tu
        roundup_time = time.time() - start_time
        print(f'cuda round up time: {roundup_time:.4f}')

        torch.cuda.synchronize()
        start_time = time.time()
        for i in range(10):
            td = _cuda_utils.double2float(t, "down")
        torch.cuda.synchronize()
        del td
        rounddown_time = time.time() - start_time
        print(f'cuda round down time: {rounddown_time:.4f}')

        del t


if __name__ == "__main__":
    if len(sys.argv) == 1:
        # Some tests. It's not possible to test them automatically because travis does not have CUDA.
        test_double2float()



class LinearBound:
    def __init__(
            self, lw=None, lb=None, uw=None, ub=None, lower=None, upper=None,
            from_input=None, x_L=None, x_U=None, offset=0, tot_dim=None):
        self.lw = lw
        self.lb = lb
        self.uw = uw
        self.ub = ub
        self.lower = lower
        self.upper = upper
        self.from_input = from_input
        self.x_L = x_L
        self.x_U = x_U
        # Offset for input variables. Used for batched forward bound
        # propagation.
        self.offset = offset
        if tot_dim is not None:
            self.tot_dim = tot_dim
        elif lw is not None:
            self.tot_dim = lw.shape[1]
        else:
            self.tot_dim = 0

    def is_single_bound(self):
        """Check whether the linear lower bound and the linear upper bound are
        the same."""
        if (self.lw is not None and self.uw is not None
                and self.lb is not None and self.ub is not None):
            return (self.lw.data_ptr() == self.uw.data_ptr()
                and self.lb.data_ptr() == self.ub.data_ptr()
                and self.x_L is not None and self.x_U is not None)
        else:
            return True



""" Base class and functions for implementing bound operators"""
from typing import Optional, List
import warnings
import torch
import torch.nn as nn
from torch import Tensor
import numpy as np

from ..perturbations import *
from ..utils import *
from ..patches import *
from ..linear_bound import LinearBound

torch._C._jit_set_profiling_executor(False)
torch._C._jit_set_profiling_mode(False)

epsilon = 1e-12


def not_implemented_op(node, func):
    message = (
        f'Function `{func}` of `{node}` is not supported yet.'
        ' Please help to open an issue at https://github.com/Verified-Intelligence/auto_LiRPA'
        ' or implement this function in auto_LiRPA/bound_ops.py'
        ' or auto_LiRPA/operators by yourself.')
    raise NotImplementedError(message)


class Interval(tuple):
    """Interval object for interval bound propagation."""

    # Subclassing tuple object so that all previous code can be reused.
    def __new__(self, lb=None, ub=None, ptb=None):
        return tuple.__new__(Interval, (lb, ub))

    def __init__(self, lb, ub, ptb=None):
        if ptb is None:
            self.ptb = None
            # `self.ptb == None` means that this interval
            # is not perturbed and it shall be treated as a constant and lb = ub.
            # To avoid mistakes, in this case the caller must make sure lb and ub are the same object.
            assert lb is ub
        else:
            if not isinstance(ptb, Perturbation):
                raise ValueError("ptb must be a Perturbation object or None. Got type {}".format(type(ptb)))
            else:
                self.ptb = ptb

    def __str__(self):
        return "({}, {}) with ptb={}".format(self[0], self[1], self.ptb)

    def __repr__(self):
        return "Interval(lb={}, ub={}, ptb={})".format(self[0], self[1], self.ptb)

    @staticmethod
    def make_interval(lb, ub, other=None):
        """Checking if the other interval is tuple, keep the perturbation."""
        if isinstance(other, Interval):
            return Interval(lb, ub, ptb=other.ptb)
        else:
            return (lb, ub)

    @staticmethod
    def get_perturbation(interval):
        """Given a tuple or Interval object, returns the norm and eps."""
        if isinstance(interval, Interval) and interval.ptb is not None:
            if isinstance(interval.ptb, PerturbationLpNorm):
                return interval.ptb.norm, interval.ptb.eps
            elif isinstance(interval.ptb, PerturbationSynonym):
                return torch.inf, 1.0
            elif isinstance(interval.ptb, PerturbationL0Norm):
                return 0, interval.ptb.eps, interval.ptb.ratio
            else:
                raise RuntimeError("get_perturbation() does not know how to handle {}".format(type(interval.ptb)))
        else:
            # Tuple object. Assuming L infinity norm lower and upper bounds.
            return torch.inf, np.nan


    @staticmethod
    def is_perturbed(interval):
        """Checking if a Interval or tuple object has perturbation enabled."""
        if isinstance(interval, Interval) and interval.ptb is None:
            return False
        else:
            return True


class Bound(nn.Module):
    r"""
    Base class for supporting the bound computation of an operator. Please see examples
    at `auto_LiRPA/operators`.

    Args:
        attr (dict): Attributes of the operator.

        inputs (list): A list of input nodes.

        output_index (int): The index in the output if the operator has multiple outputs. Usually output_index=0.

        options (dict): Bound options.

    Be sure to run `super().__init__(attr, inputs, output_index, options, device)`
    first in the `__init__` function.
    """

    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__()
        attr = {} if attr is None else attr
        inputs = [] if inputs is None else inputs
        options = {} if options is None else options
        self.name: Optional[str] = None
        self.output_name = []
        self.device = attr.get('device')
        self.attr, self.inputs, self.output_index, self.options = \
            attr, inputs, output_index, options
        self.forward_value = None
        self.output_shape = None
        self.from_input = False
        self.bounded = False
        self.IBP_rets = None
        self.requires_input_bounds = []
        # If True, when we are computing intermediate bounds for these ops,
        # we simply use IBP to propagate bounds from its input nodes
        # instead of CROWN. Currently only operators with a single input can be
        # supported.
        self.ibp_intermediate = False
        self.splittable = False
        # Determine if this node has a perturbed output or not. The function BoundedModule._mark_perturbed_nodes() will set this property.
        self.perturbed = False
        self.never_perturbed = False
        if options is not None and 'loss_fusion' in options:
            self.loss_fusion = options['loss_fusion']
        else:
            self.loss_fusion = False
        self.options = options
        # Use `default_interval_propagate`
        self.use_default_ibp = False
        # If set to true, the backward bound output of this node is 0.
        self.zero_backward_coeffs_l = False
        self.zero_backward_coeffs_u = False
        # If set to true, the A matrix accumulated on this node is 0.
        self.zero_lA_mtx = False
        self.zero_uA_mtx = False
        self.patches_start = False
        self.alpha_beta_update_mask = None

    def __repr__(self, attrs=None):
        inputs = ', '.join([node.name for node in self.inputs])
        ret = (f'{self.__class__.__name__}(name={self.name}, '
                f'inputs=[{inputs}], perturbed={self.perturbed}')
        if attrs is not None:
            for k, v in attrs.items():
                ret += f', {k}={v}'
        ret += ')'
        return ret

    def is_input_perturbed(self, i=0):
        r"""Check if the i-th input is with perturbation or not."""
        return i < len(self.inputs) and self.inputs[i].perturbed

    def clear(self):
        """ Clear attributes when there is a new input to the network"""
        pass

    @property
    def input_name(self):
        return [node.name for node in self.inputs]

    def forward(self, *x):
        r"""
        Function for standard/clean forward.

        Args:
            x: A list of input values. The length of the list is equal to the number of input nodes.

        Returns:
            output (Tensor): The standard/clean output of this node.
        """
        return not_implemented_op(self, 'forward')

    def interval_propagate(self, *v):
        r"""
        Function for interval bound propagation (IBP) computation.

        There is a default function `self.default_interval_propagate(*v)` in the base class,
        which can be used if the operator is *monotonic*. To use it, set `self.use_default_ibp = True`
        in the `__init__` function, and the implementation of this function can be skipped.

        Args:
            v: A list of the interval bound of input nodes.
            Generally, for each element `v[i]`, `v[i][0]` is the lower interval bound,
            and `v[i][1]` is the upper interval bound.

        Returns:
            bound: The interval bound of this node, in a same format as v[i].
        """
        if self.use_default_ibp or self.never_perturbed:
            return self.default_interval_propagate(*v)
        else:
            return not_implemented_op(self, 'interval_propagate')

    def default_interval_propagate(self, *v):
        """Default IBP using the forward function.

        For unary monotonous functions or functions for altering shapes only
        but not values.
        """
        if len(v) == 0:
            return Interval.make_interval(self.forward(), self.forward())
        else:
            if len(v) > 1:
                for i in range(1, len(v)):
                    assert not self.is_input_perturbed(i)
            return Interval.make_interval(
                self.forward(v[0][0], *[vv[0] for vv in v[1:]]),
                self.forward(v[0][1], *[vv[0] for vv in v[1:]]), v[0])

    def bound_forward(self, dim_in, *x):
        r"""
        Function for forward mode bound propagation.

        Forward mode LiRPA computs a `LinearBound`
        instance representing the linear bound for each involved node.
        Major attributes of `LinearBound` include
        `lw`, `uw`, `lb`, `ub`, `lower`, and `upper`.

        `lw` and `uw` are coefficients of linear bounds w.r.t. model input.
        Their shape is `(batch_size, dim_in, *standard_shape)`,
        where `dim_in` is the total dimension of perturbed input nodes of the model,
        and `standard_shape` is the shape of the standard/clean output.
        `lb` and `ub` are bias terms of linear bounds, and their shape is equal
        to the shape of standard/clean output.
        `lower` and `upper` are concretized lower and upper bounds that will be
        computed later in BoundedModule.

        Args:
            dim_in (int): Total dimension of perturbed input nodes of the model.

            x: A list of the linear bound of input nodes. Each element in x is a `LinearBound` instance.

        Returns:
            bound (LinearBound): The linear bound of this node.
        """
        return not_implemented_op(self, 'bound_forward')

    def bound_dynamic_forward(self, *x, max_dim=None, offset=0):
        raise NotImplementedError(f'bound_dynamic_forward is not implemented for {self}.')

    def bound_backward(self, last_lA, last_uA, *x, **kwargs):
        r"""
        Function for backward mode bound propagation.

        Args:
            last_lA (Tensor): `A` matrix for lower bound computation propagated to this node. It can be `None` if lower bound is not needed.

            last_uA (Tensor): `A` matrix for upper bound computation propagated to this node. It can be `None` if upper bound is not needed.

            x: A list of input nodes, with x[i].lower and x[i].upper that can be used as pre-activation bounds.

        Returns:
            A: A list of A matrices for the input nodes. Each element is a tuple (lA, uA).

            lbias (Tensor): The bias term for lower bound computation, introduced by the linear relaxation of this node. .

            ubias (Tensor): The bias term for upper bound computation, introduced by the linear relaxation of this node.
        """
        return not_implemented_op(self, 'bound_backward')

    def broadcast_backward(self, A, x):
        shape = x.output_shape

        if isinstance(A, Tensor):
            if x.batch_dim == -1:
                # final shape of input
                shape = torch.Size([A.shape[1]] + list(shape))
                dims = []
                cnt_sum = A.ndim - len(shape) - 1
                for i in range(2, A.ndim): # merge the output dimensions?
                    if cnt_sum > 0:
                        dims.append(i)
                        cnt_sum -= 1
                if dims:
                    A = torch.sum(A, dim=dims)
            else:
                dims = list(range(1, A.ndim - len(shape)))
                if dims:
                    A = torch.sum(A, dim=dims)
            dims = []
            for i in range(1, len(shape)):
                # Skip the batch dimension.
                # FIXME (05/11/2022): the following condition is not always correct.
                # We should not rely on checking dimension is "1" or not.
                if shape[i] == 1 and A.shape[i + 1] != 1:
                    dims.append(i + 1)
            if dims:
                A = torch.sum(A, dim=dims, keepdim=True)
            assert A.shape[2:] == shape[1:]  # skip the spec and batch dimension.
        else:
            pass
        return A

    def build_gradient_node(self, grad_upstream):
        r"""
        Function for building the gradient node to bound the Jacobian.

        Args:
            grad_upstream: Upstream gradient in the gradient back-propagation.

        Returns:
            module_grad (torch.nn.Module): Gradient node.

            grad_input (list): Inputs to the gradient node. Values do not
            matter. We only want the shapes.

            grad_extra_nodes (list): Extra nodes needed for the gradient.
        """
        return not_implemented_op(self, 'build_gradient_node')

    def get_bias(self, A, bias):
        if A is None:
            return 0
        if not Benchmarking:
            assert not isnan(A)
            assert not isnan(bias)
        if torch.isinf(bias).any():
            warnings.warn('There is an inf value in the bias of LiRPA bounds.')

        if isinstance(A, Tensor):
            if self.batch_dim != -1:
                bias_new = torch.einsum('sb...,b...->sb', A, bias)
            else:
                bias_new = torch.einsum('sb...,...->sb', A, bias)
            if isnan(bias_new):
                # NaN can be caused by 0 * inf, if 0 appears in `A` and inf appears in `bias`.
                # Force the whole bias to be 0, to avoid gradient issues.
                # FIXME maybe find a more robust solution.
                return 0
            else:
                # FIXME (09/17): handle the case for pieces.unstable_idx.
                return bias_new
        elif isinstance(A, eyeC):
            batch_size = A.shape[1]
            if self.batch_dim != -1:
                return bias.reshape(batch_size, -1).t()
            else:
                return bias.reshape(-1).unsqueeze(-1).repeat(1, batch_size)
        elif type(A) == Patches:
            # the shape of A.patches is [batch, L, out_c, in_c, K, K]
            if self.batch_dim != -1:
                # Input A patches has shape (spec, batch, out_h, out_w, in_c, H, W) or (unstable_size, batch, in_c, H, W).
                patches = A.patches
                # Here the size of bias is [batch_size, out_h, out_w, in_c, H, W]
                bias = inplace_unfold(bias, kernel_size=A.patches.shape[-2:], stride=A.stride, padding=A.padding, inserted_zeros=A.inserted_zeros, output_padding=A.output_padding)
                if A.unstable_idx is not None:
                    # Sparse bias has shape [unstable_size, batch_size, in_c, H, W]. No need to select over the out_c dimension.
                    bias = bias[:, A.unstable_idx[1], A.unstable_idx[2]]
                    # bias_new has shape (unstable_size, batch).
                    bias_new = torch.einsum('bschw,sbchw->sb', bias, patches)
                else:
                    # Sum over the in_c, H, W dimension. Use torch.einsum() to save memory, equal to:
                    # bias_new = (bias * patches).sum(-1,-2,-3).transpose(-2, -1)
                    # bias_new has shape (spec, batch, out_h, out_w).
                    bias_new = torch.einsum('bijchw,sbijchw->sbij', bias, patches)
            else:
                # Similar to BoundConstant. (BoundConstant does not have batch_dim).
                # FIXME (09/16): bias size is different for BoundConstant. We should use the same size!
                patches = A.patches
                bias_new = torch.sum(patches, dim=(-1, -2, -3)) * bias.to(self.device)
                # Return shape is (spec, batch, out_h, out_w) or (unstable_size, batch).
                return bias_new
            return bias_new
        else:
            return NotImplementedError()

    def make_axis_non_negative(self, axis, shape='input'):
        if isinstance(axis, (tuple, list)):
            return tuple([self.make_axis_non_negative(item, shape)
                          for item in axis])
        if shape == 'input':
            shape = self.input_shape
        elif shape == 'output':
            shape = self.output_shape
        else:
            assert isinstance(shape, torch.Size)
        if axis < 0:
            return axis + len(shape)
        else:
            return axis

    def update_requires_input_bounds(self):
        """Update requires_input_bounds.

        This function is called once we know if the input nodesare perturbed.
        """
        pass

    def clamp_interim_bounds(self):
        """Clamp intermediate bounds."""
        pass

    def check_constraint_available(self, node, flag=False):
        if hasattr(node, 'cstr_interval'):
            flag = True
        for n in node.inputs:
            if not n.from_input:
                flag = flag or self.check_constraint_available(n, flag)
        return flag

    def _ibp_constraint(self, node, delete_bounds_after_use=False):
        def _delete_unused_bounds(node_list):
            """Delete bounds from input layers after use to save memory. Used when
            sparse_intermediate_bounds_with_ibp is true."""
            if delete_bounds_after_use:
                for n in node_list:
                    del n.cstr_interval
                    del n.cstr_lower
                    del n.cstr_upper

        if not node.perturbed and hasattr(node, 'forward_value'):
            node.cstr_lower, node.cstr_upper = node.cstr_interval = (
                node.forward_value, node.forward_value)

        to_be_deleted_bounds = []
        if not hasattr(node, 'cstr_interval'):
            for n in node.inputs:
                if not hasattr(n, 'cstr_interval'):
                    # Node n does not have interval bounds; we must compute it.
                    self._ibp_constraint(
                        n, delete_bounds_after_use=delete_bounds_after_use)
                    to_be_deleted_bounds.append(n)
            inp = [n_pre.cstr_interval for n_pre in node.inputs]
            node.cstr_interval = node.interval_propagate(*inp)

            node.cstr_lower, node.cstr_upper = node.cstr_interval
            if isinstance(node.cstr_lower, torch.Size):
                node.cstr_lower = torch.tensor(node.cstr_lower)
                node.cstr_interval = (node.cstr_lower, node.cstr_upper)
            if isinstance(node.cstr_upper, torch.Size):
                node.cstr_upper = torch.tensor(node.cstr_upper)
                node.cstr_interval = (node.cstr_lower, node.cstr_upper)

        if hasattr(node, 'lower'):
            node.lower = torch.where(node.lower >= node.cstr_lower, node.lower,
                            node.cstr_lower)
            node.upper = torch.where(node.upper <= node.cstr_upper, node.upper,
                            node.cstr_upper)
            node.interval = (node.lower, node.upper)

        _delete_unused_bounds(to_be_deleted_bounds)
        return node.cstr_interval

    def _check_weight_perturbation(self):
        weight_perturbation = False
        for n in self.inputs[1:]:
            if hasattr(n, 'perturbation'):
                if n.perturbation is not None:
                    weight_perturbation = True
        if weight_perturbation:
            self.requires_input_bounds = list(range(len(self.inputs)))
        else:
            self.requires_input_bounds = []
        return weight_perturbation

    def non_deter_wrapper(self, op, *args, **kwargs):
        """Some operations are non-deterministic and deterministic mode will fail.
        So we temporary disable it."""
        if self.options.get('deterministic', False):
            torch.use_deterministic_algorithms(False)
        ret = op(*args, **kwargs)
        if self.options.get('deterministic', False):
            torch.use_deterministic_algorithms(True)
        return ret

    def non_deter_scatter_add(self, *args, **kwargs):
        return self.non_deter_wrapper(torch.scatter_add, *args, **kwargs)

    def non_deter_index_select(self, *args, **kwargs):
        return self.non_deter_wrapper(torch.index_select, *args, **kwargs)



import gurobipy as grb
from example.test_model import extract_instance
from verifier.verifier import Verifier 
import torch.nn as nn
import torch

class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.layers = nn.Sequential(*[
            nn.Linear(5, 4),
            nn.ReLU(),
            nn.Linear(4, 3),
            nn.ReLU(),
            nn.Linear(3, 2),
            nn.ReLU(),
            nn.Linear(2, 5),
        ])
        
        self.layers[0].weight.data = torch.tensor([
            [1., -1., 2., -2., 3.],
            [1., 2., -3., -4., 5.],
            [-1., -2., 1., -1., -3.],
            [-1., 1., 1., -1., 4.],
        ])
        self.layers[0].bias.data = torch.tensor(
            [1., -1., 2., -4.],
        )
            
        self.layers[2].weight.data = torch.tensor([
            [1., -1., -2., -3.],
            [1., 2., -4., 5.],
            [-1., 5., -1., 3.],
        ])
        self.layers[2].bias.data = torch.tensor(
            [1., -1., 2.],
        )
        
        self.layers[4].weight.data = torch.tensor([
            [1., -2., 3.],
            [2., -4., -5.],
        ])
        self.layers[4].bias.data = torch.tensor(
            [1., -1.],
        )
        
        self.layers[6].weight.data = torch.tensor([
            [1., -2.],
            [-4., -5.],
            [-1., 5.],
            [1., -3.],
            [2., 3.],
        ])
        self.layers[6].bias.data = torch.tensor(
            [1., -1., -2., -3., 1.],
        )
        
        
        
    @torch.no_grad()
    def forward(self, x):
        for layer in self.layers:
            # print(x)
            x = layer(x)
            # print(x)
            # print()
        return x
        # return self.layers(x)



if __name__ == "__main__":
    # model = grb.Model()
    # x0 = model.addVar(name='x0', lb=-1, ub=1)
    # x1 = model.addVar(name='x1', lb=-1, ub=2)
    
    # model.addConstr(x0 + x1 <= 0)
    # model.addConstr(x0 - x1 >= -2)
    
    # a = model.addVar(name='new_var_a', lb=-10, ub=20)
    # b = model.addVar(name='new_var_b', lb=-10, ub=20)
    # model.addConstr(x0 + 2 * x1 >= a)
    # model.addConstr(2*x0 - 3 * x1 == b)
    # model.update()
    # print(model)
    
    # model.remove(a)
    # model.update()
    # print(model)
    # model.write('test_gurobi.lp')
    
    net_path = 'example/gurobi.onnx'
    vnnlib_path = 'example/prop_3.vnnlib'
    device = 'cpu'
    # torch.manual_seed(1)
    
    net = Net()
    x = torch.randn(1, 5)
    print(net(x))
    
    net.eval()
    torch.onnx.export(
        net,
        x,
        net_path,
        verbose=False,
        opset_version=12,
    )
    
    exit()
    
    
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)
    
    verifier = Verifier(
        net=model, 
        input_shape=input_shape, 
        batch=1,
        device=device,
    )
    
    status = verifier.verify(objectives)


import torch
from torch import nn


class GatherND(nn.Module):
    def __init__(self, batch_dims=0):
        if batch_dims != 0:
            raise NotImplementedError(
                f"GatherND for batch_dims={batch_dims} not implemented."
            )
        self.batch_dims = batch_dims
        super().__init__()

    def forward(self, data: torch.Tensor, indices: torch.Tensor):
        orig_shape = list(indices.shape)
        num_samples = torch.prod(torch.tensor(orig_shape[:-1]))
        m = orig_shape[-1]
        n = len(data.shape)

        if m > n:
            raise ValueError(
                f"The last dimension of indices must be <= the rank of data."
                f"Got indices:{indices.shape}, data:{data.shape}. {m} > {n}"
            )
        out_shape = orig_shape[:-1] + list(data.shape)[m:]

        indices = indices.reshape((num_samples, m)).transpose(0, 1)
        indices = torch.split(indices, 1, 0)
        output = data[indices]  # (num_samples, ...)
        return output.reshape(out_shape).contiguous()



from torch.optim import Optimizer
from beartype import beartype
import typing
import torch
import math


class AdamClipping(Optimizer):

    r"""Implements Adam algorithm, with per-parameter gradient clipping.
    The function is from PyTorch source code.

    It has been proposed in `Adam: A Method for Stochastic Optimization`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    # @beartype
    def __init__(self: 'AdamClipping', params: list[torch.Tensor], lr: float = 1e-3, betas: tuple[float, float] = (0.9, 0.999),
                 eps: float = 1e-8, weight_decay: float = 0.0, amsgrad: bool = False) -> None:
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, amsgrad=amsgrad)
        super().__init__(params, defaults)

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)


    # # @beartype
    @staticmethod
    @torch.no_grad()
    @torch.jit.script
    def _clip_update(exp_avg: torch.Tensor, denom: torch.Tensor, step_size: float, clipping_step_eps: float, lower_limit: torch.Tensor, upper_limit: torch.Tensor, p: torch.Tensor) -> None:
        # Compute the Adam update.
        update = exp_avg / denom * step_size
        # update = p.grad
        # Linf norm, scale according to sign.
        scaled_update = torch.sign(update) * clipping_step_eps
        # Apply the update.
        d = p.data + scaled_update
        # Avoid out-of-boundary updates.
        d = torch.max(torch.min(d, upper_limit), lower_limit)
        p.copy_(d)

    # @beartype
    @torch.no_grad()
    def step(self: 'AdamClipping', clipping: bool = False, lower_limit: torch.Tensor | None = None, upper_limit: torch.Tensor | None = None, sign: int | None = None) -> None:
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """

        # Currently we only deal with 1 parameter group.
        assert len(self.param_groups) == 1
        for group in self.param_groups:
            for i, p in enumerate(group['params']):
                if p.grad is None:
                    continue
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')
                amsgrad = group['amsgrad']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                if group['weight_decay'] != 0:
                    grad = grad.add(p, alpha=group['weight_decay'])

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = (max_exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                else:
                    denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])

                step_size = group['lr'] / bias_correction1

                if clipping:
                    assert sign == 1  # gradient ascent for adversarial attacks.
                    self._clip_update(exp_avg, denom, step_size, step_size, lower_limit, upper_limit, p)
                else:
                    # No clipping. Original Adam update.
                    p.addcdiv_(exp_avg, denom, value=-step_size)

        return None



from beartype import beartype
import torch

from ..network.read_onnx import inference_onnx

# @beartype
def get_adv_string(inputs: torch.Tensor, net_path: str, is_nhwc: bool = False) -> str:
    if is_nhwc:
        assert inputs.ndim == 4
        inputs = inputs.permute(0, 2, 3, 1)
    x = inputs.detach().cpu().float().numpy()
    y = inference_onnx(net_path, x)[0]
    # flatten
    x = x.flatten()
    y = y.flatten()
    # export
    string_x = [f'(X_{i} {x[i]})' for i in range(len(x))]
    string_y = [f'(Y_{i} {y[i]})' for i in range(len(y))]
    string = '\n '.join(string_x + string_y)
    return f"({string})"



from torch import nn
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.pooling import _MaxPoolNd
from .operations import (
    BatchNormWrapper,
    InstanceNormWrapper,
    Loop,
    LSTMWrapper,
    Split,
    TopK,
)


COMPOSITE_LAYERS = (nn.Sequential,)
MULTIOUTPUT_LAYERS = (Loop, LSTMWrapper, Split, TopK)
STANDARD_LAYERS = (
    _ConvNd,
    BatchNormWrapper,
    InstanceNormWrapper,
    LSTMWrapper,
    nn.Linear,
)



from beartype import beartype
import gurobipy as grb
import numpy as np
import typing
import torch
import math
import copy
import os

from ..auto_LiRPA.perturbations import PerturbationLpNorm
from ..auto_LiRPA import BoundedTensor

from ..util.misc.check import check_solution
from ..util.misc.logger import logger
from .params import get_branching_opt_params

if typing.TYPE_CHECKING:
    from .. import abstractor

def update_refined_beta(self: 'abstractor.abstractor.NetworkAbstractor', betas, batch):
    pass

# @beartype
def new_input(self: 'abstractor.abstractor.NetworkAbstractor', x_L: torch.Tensor, x_U: torch.Tensor) -> BoundedTensor:
    if os.environ.get('NEURALSAT_ASSERT'):
        assert torch.all(x_L <= x_U)
    return BoundedTensor(x_L, PerturbationLpNorm(x_L=x_L, x_U=x_U)).to(self.device)


# @beartype
def new_slopes(slopes: dict, keep_name: str) -> dict:
    new_slope = {}
    for relu_layer, alphas in slopes.items():
        new_slope[relu_layer] = {}
        if keep_name in alphas:
            new_slope[relu_layer][keep_name] = alphas[keep_name]
    return new_slope


# @beartype
def _to_device(tensor: torch.Tensor, device: str = 'cpu', half: bool = False) -> torch.Tensor:
    assert device in ['cpu', 'cuda'] and isinstance(half, bool)
    if half:
        tensor = tensor.half()
    if device:
        tensor = tensor.to(device)
    return tensor


# @beartype
def get_slope(self: 'abstractor.abstractor.NetworkAbstractor', half=True, device='cpu') -> dict:
    if len(self.net.perturbed_optimizable_activations) == 0:
        return {}
    slopes = {
        m.name: {
            node_name: _to_device(alpha, device=device, half=half) for (node_name, alpha) in m.alpha.items()
        } for m in self.net.perturbed_optimizable_activations
    }
    return slopes


# @beartype
def set_slope(self: 'abstractor.abstractor.NetworkAbstractor', slope: dict) -> None:
    for m in self.net.perturbed_optimizable_activations:
        for node_name in list(m.alpha.keys()):
            if node_name in slope[m.name]:
                slope_len = slope[m.name][node_name].size(2)
                if slope_len > 0:
                    m.alpha[node_name] = slope[m.name][node_name]
                    m.alpha[node_name] = m.alpha[node_name].repeat(1, 1, 2, *([1] * (m.alpha[node_name].ndim - 3))).detach().requires_grad_() # 2 * batch
                    # print('setting alpha:', m.name, node_name, m.alpha[node_name].shape, m.alpha[node_name].dtype, m.alpha[node_name].sum().item())
            else:
                # do not use alphas
                del m.alpha[node_name]


# @beartype
def get_hidden_bounds(self: 'abstractor.abstractor.NetworkAbstractor', output_lbs: torch.Tensor, device: str = 'cpu') -> tuple[dict, dict]:
    lower_bounds, upper_bounds = {}, {}
    output_ubs = output_lbs + torch.inf

    # get hidden bounds
    for layer in list(set(self.net.layers_requiring_bounds + self.net.split_nodes)):
        lower_bounds[layer.name] = _to_device(layer.lower.detach(), device=device)
        upper_bounds[layer.name] = _to_device(layer.upper.detach(), device=device)

    # add output bounds
    lower_bounds[self.net.final_name] = _to_device(output_lbs.flatten(1).detach(), device=device)
    upper_bounds[self.net.final_name] = _to_device(output_ubs.flatten(1).detach(), device=device)

    assert len(list(set([_.shape[0] for _ in lower_bounds.values()]))) == 1, print([_.shape[0] for _ in lower_bounds.values()])
    assert len(list(set([_.shape[0] for _ in upper_bounds.values()]))) == 1, print([_.shape[0] for _ in upper_bounds.values()])

    return lower_bounds, upper_bounds


# @beartype
def get_lAs(self: 'abstractor.abstractor.NetworkAbstractor', size: int | None = None, device: str = 'cpu') -> dict:
    lAs = {}
    list_nodes = [n for n in self.net.nodes() if n.name == self.net.input_name[0]] if self.input_split else self.net.get_splittable_activations()
    for node in list_nodes:
        lA = getattr(node, 'lA', None)
        if lA is None:
            continue
        lAs[node.name] = _to_device(lA.transpose(0, 1), device=device)
    return lAs


# @beartype
def get_beta(self: 'abstractor.abstractor.NetworkAbstractor', num_splits: list[dict], device: str = 'cpu') -> list:
    ret = []
    for i in range(len(num_splits)):
        betas = {k: _to_device(self.net[k].sparse_betas[0].val[i, :num_splits[i][k]], device=device) for k in num_splits[i]}
        ret.append(betas)
    return ret


# @beartype
def reset_beta(self: 'abstractor.abstractor.NetworkAbstractor', batch: int, max_splits_per_layer: dict, betas: list | None = None, bias: bool = False) -> None:
    for layer_name in max_splits_per_layer:
        layer = self.net[layer_name]
        start_nodes = []
        for act in self.net.split_activations[layer_name]:
            start_nodes.extend(list(act[0].alpha.keys()))
        shape = (batch, max_splits_per_layer[layer_name])
        if betas is not None and betas[0] is not None and layer_name in betas[0]:
            betas_ = [(betas[bi][layer_name] if betas[bi] is not None else None) for bi in range(batch)]
        else:
            betas_ = [None for _ in range(batch)]
        # set betas
        self.net.reset_beta(layer, shape, betas_, bias=bias, start_nodes=list(set(start_nodes)))


# @beartype
def _copy_history(history: dict) -> dict:
    ret = {}
    for k, v in history.items():
        if isinstance(v[0], torch.Tensor):
            ret[k] = v
        elif isinstance(v[0], list):
            ret[k] = tuple(v[i].copy() for i in range(len(v)))
        else:
            ret[k] = tuple(copy.deepcopy(v[i]) for i in range(len(v)))
    return ret


# @beartype
def _append_tensor(tensor: torch.Tensor | list, value: int | float | np.int64, dtype: torch.dtype = torch.float32) -> torch.Tensor:
    if not isinstance(tensor, torch.Tensor):
        tensor = torch.tensor(tensor, dtype=dtype)
    size = len(tensor)
    res = torch.empty(size=(size+1,), dtype=dtype)
    res[:size] = tensor
    res[-1] = value
    return res


# @beartype
def update_histories(self: 'abstractor.abstractor.NetworkAbstractor', histories: list[dict], decisions: list[list]) -> list[dict]:
    double_histories = []
    batch = len(decisions)

    # double the histories
    for _ in range(2):
        for h in histories:
            double_histories.append(_copy_history(h))

    # add new decisions to histories
    for i, h in enumerate(double_histories):
        bi = i % batch
        n_name, n_id, n_point = decisions[bi]
        # FIXME: check repeated decisions
        loc = _append_tensor(h[n_name][0], n_id, dtype=torch.long)
        sign = _append_tensor(h[n_name][1], +1 if i < batch else -1)
        beta = _append_tensor(h[n_name][2], n_point)
        h[n_name] = (loc, sign, beta)

    return double_histories


# @beartype
def set_beta(self: 'abstractor.abstractor.NetworkAbstractor', betas: list, histories: list[dict]) -> list[dict]:
    batch = len(histories)
    splits_per_example = []
    max_splits_per_layer = {}

    for bi in range(batch):
        splits_per_example.append({})
        for k, v in histories[bi].items():
            splits_per_example[bi][k] = len(v[0])
            max_splits_per_layer[k] = max(max_splits_per_layer.get(k, 0), splits_per_example[bi][k])

    # set old betas
    self.reset_beta(
        betas=betas,
        max_splits_per_layer=max_splits_per_layer,
        batch=batch,
        bias=None in self.split_points,
    )

    # set new betas
    for node in self.net.split_nodes:
        if node.sparse_betas is None:
            continue
        sparse_betas = node.sparse_betas if isinstance(node.sparse_betas, list) else node.sparse_betas.values()
        for sparse_beta in sparse_betas:
            sparse_beta.apply_splits(histories, node.name)

    return splits_per_example


# @beartype
@torch.no_grad()
def hidden_split_idx(self: 'abstractor.abstractor.NetworkAbstractor', lower_bounds: dict, upper_bounds: dict,
                     decisions: list[list]) -> dict:
    batch = len(decisions)
    splitting_indices_batch = {k: [] for k in lower_bounds}
    splitting_indices_neuron = {k: [] for k in lower_bounds}
    splitting_points = {k: [] for k in lower_bounds}

    for i in range(batch):
        n_name, n_id, n_point = decisions[i]
        splitting_indices_batch[n_name].append(i)
        splitting_indices_neuron[n_name].append(n_id)
        splitting_points[n_name].append(n_point)

    # convert to tensor
    splitting_indices_batch = {k: torch.as_tensor(v).to(device=self.device, non_blocking=True) for k, v in splitting_indices_batch.items()}
    splitting_indices_neuron = {k: torch.as_tensor(v).to(device=self.device, non_blocking=True) for k, v in splitting_indices_neuron.items()}
    splitting_points = {k: torch.as_tensor(v).to(device=self.device, non_blocking=True) for k, v in splitting_points.items()}

    # 2 * batch
    double_upper_bounds = {k: torch.cat([v, v], dim=0) for k, v in upper_bounds.items()}
    double_lower_bounds = {k: torch.cat([v, v], dim=0) for k, v in lower_bounds.items()}

    # construct new hidden bounds
    new_intermediate_layer_bounds = {}
    for key in double_lower_bounds:
        assert len(double_lower_bounds[key]) == len(double_upper_bounds[key]) == 2 * batch
        if len(splitting_indices_batch[key]):
            # set 1st half (set lower)
            double_lower_bounds[key].view(2 * batch, -1)[splitting_indices_batch[key], splitting_indices_neuron[key]] = splitting_points[key]
            # set 2nd half (set upper)
            double_upper_bounds[key].view(2 * batch, -1)[splitting_indices_batch[key] + batch, splitting_indices_neuron[key]] = splitting_points[key]
            if os.environ.get('NEURALSAT_ASSERT'):
                assert torch.all(double_lower_bounds[key] <= double_upper_bounds[key])
        new_intermediate_layer_bounds[key] = [double_lower_bounds[key], double_upper_bounds[key]]

    assert all([_[0].shape[0] == _[1].shape[0] == 2 * batch for _ in new_intermediate_layer_bounds.values()])
    return new_intermediate_layer_bounds


# @beartype
@torch.no_grad()
def input_split_idx(self: 'abstractor.abstractor.NetworkAbstractor', input_lowers: torch.Tensor, input_uppers: torch.Tensor,
                    split_idx: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    input_lowers = input_lowers.flatten(1)
    input_uppers = input_uppers.flatten(1)

    input_lowers_cp = input_lowers.clone()
    input_uppers_cp = input_uppers.clone()

    indices = torch.arange(input_lowers_cp.shape[0])
    idx = split_idx[:, 0].long()

    input_lowers_cp_tmp = input_lowers_cp.clone()
    input_uppers_cp_tmp = input_uppers_cp.clone()

    split_value = (input_lowers_cp[indices, idx] + input_uppers_cp[indices, idx]) / 2

    input_lowers_cp[indices, idx] = split_value
    input_uppers_cp_tmp[indices, idx] = split_value

    input_lowers_cp = torch.cat([input_lowers_cp, input_lowers_cp_tmp])
    input_uppers_cp = torch.cat([input_uppers_cp, input_uppers_cp_tmp])

    new_input_lowers = input_lowers_cp.reshape(-1, *self.input_shape[1:])
    new_input_uppers = input_uppers_cp.reshape(-1, *self.input_shape[1:])

    return new_input_lowers, new_input_uppers


# @beartype
def build_lp_solver(self: 'abstractor.abstractor.NetworkAbstractor', model_type: str,
                    input_lower: torch.Tensor, input_upper: torch.Tensor,
                    c: torch.Tensor, rhs: torch.Tensor | None = None,
                    refine: bool = True, intermediate_layer_bounds: dict | None = None,
                    timeout: float | None = None, timeout_per_neuron: float | None = None) -> None:
    assert model_type in ['lp', 'mip']
    # delete old LP model
    self.net._reset_solver_vars(self.net.final_node())
    if hasattr(self.net, 'model'):
        del self.net.model

    # gurobi solver
    self.net.model = grb.Model(model_type)
    self.net.model.setParam('OutputFlag', False)
    self.net.model.setParam("FeasibilityTol", 1e-5)
    # self.net.model.setParam('TimeLimit', timeout)
    if model_type == 'mip':
        self.net.model.setParam('MIPGap', 1e-2)  # Relative gap between lower and upper objective bound
        self.net.model.setParam('MIPGapAbs', 1e-2)  # Absolute gap between lower and upper objective bound

    # create new inputs
    new_x = self.new_input(input_lower, input_upper)

    # forward to recompute hidden bounds
    self.net.set_bound_opts(get_branching_opt_params())
    self.net.init_alpha(x=(new_x,), c=c)

    lb, _ = self.net.compute_bounds(x=(new_x,), C=c, method="backward", reference_bounds=intermediate_layer_bounds)
    if rhs is not None:
        if (lb > rhs).all():
            return None

    # build solver
    self.net.build_solver_module(
        x=(new_x,),
        # intermediate_layer_bounds=intermediate_layer_bounds,
        C=c,
        final_node_name=self.net.final_name,
        model_type=model_type,
        timeout=timeout,
        timeout_per_neuron=timeout_per_neuron,
        refine=refine,
    )
    self.net.model.update()
    self.last_c_lp = c
    self.last_input_lower = input_lower.clone()
    self.last_input_upper = input_upper.clone()


# @beartype
def solve_full_assignment(self: 'abstractor.abstractor.NetworkAbstractor', input_lower: torch.Tensor, input_upper: torch.Tensor,
                          lower_bounds: dict, upper_bounds: dict, c: torch.Tensor, rhs: torch.Tensor) -> tuple[bool, torch.Tensor | None]:
    logger.debug('Full assignment')
    tmp_model = self.net.model.copy()
    tmp_model.update()

    # TODO: assert all activation layers are ReLU
    pre_relu_layer_names = [relu_layer.inputs[0].name for relu_layer in self.net.relus]
    relu_layer_names = [relu_layer.name for relu_layer in self.net.relus]

    for relu_idx, (pre_relu_name, relu_name) in enumerate(zip(pre_relu_layer_names, relu_layer_names)):
        lbs = lower_bounds[pre_relu_name].reshape(-1)
        ubs = upper_bounds[pre_relu_name].reshape(-1)
        for neuron_idx in range(lbs.shape[0]):
            pre_var = tmp_model.getVarByName(f"lay{pre_relu_name}_{neuron_idx}")
            pre_var.lb = pre_lb = lbs[neuron_idx]
            pre_var.ub = pre_ub = ubs[neuron_idx]
            var = tmp_model.getVarByName(f"ReLU{relu_name}_{neuron_idx}")
            # var is None if originally stable
            if var is not None:
                if pre_ub >= pre_lb >= 0:
                    var.lb = pre_lb
                    var.ub = pre_ub
                    tmp_model.addConstr(pre_var == var)
                elif pre_lb <= pre_ub <= 0:
                    var.lb = 0
                    var.ub = 0
                else:
                    raise ValueError(f'Exists unstable neuron at index [{relu_idx}][{neuron_idx}]: lb={pre_lb} ub={pre_ub}')

    tmp_model.update()

    feasible = True
    adv = None
    output_vars = self.net.final_node().solver_vars
    assert len(output_vars) == len(rhs), f"out shape not matching! {len(output_vars)} {len(rhs)}"
    for out_idx in range(len(output_vars)):
        objective_var = tmp_model.getVarByName(output_vars[out_idx].VarName)
        tmp_model.setObjective(objective_var, grb.GRB.MINIMIZE)
        tmp_model.update()
        tmp_model.optimize()

        if tmp_model.status == 2:
            # print("Gurobi all node split: feasible!")
            output_lb = objective_var.X
        else:
            # print(f"Gurobi all node split: infeasible! Model status {tmp_model.status}")
            output_lb = float('inf')

        if output_lb > rhs[out_idx]:
            feasible = False
            adv = None
            break

        input_vars = [tmp_model.getVarByName(f'inp_{dim}') for dim in range(math.prod(self.input_shape))]
        adv = torch.tensor([var.X for var in input_vars], device=self.device).view(self.input_shape)
        if check_solution(net=self.pytorch_model, adv=adv, cs=c, rhs=rhs, data_min=input_lower, data_max=input_upper):
            return True, adv

    del tmp_model
    return feasible, adv



# @beartype
def compute_stability(self: 'abstractor.abstractor.NetworkAbstractor', objective):
    cs = objective.cs.to(self.device)
    rhs = objective.rhs.to(self.device)

    # input property
    if not torch.allclose(objective.lower_bounds.mean(dim=0), objective.lower_bounds[0], 1e-5, 1e-5):
        input_lowers = objective.lower_bounds.view(-1, *self.input_shape[1:]).to(self.device)
        input_uppers = objective.upper_bounds.view(-1, *self.input_shape[1:]).to(self.device)
    else:
        input_lowers = objective.lower_bounds[0:1].to(self.device)
        input_uppers = objective.upper_bounds[0:1].to(self.device)

    x = self.new_input(x_L=input_lowers, x_U=input_uppers)

    assert self.method in ['forward', 'backward']
    with torch.no_grad():
        lb, _ = self.net.compute_bounds(
            x=(x,),
            C=cs,
            method=self.method,
        )
        lower_bounds, upper_bounds = [], []
        for node in self.net.relus:
            lower_bounds.append(node.inputs[0].lower)
            upper_bounds.append(node.inputs[0].upper)

    n_unstable = sum([
        torch.logical_and(lower_bounds[j] < 0, upper_bounds[j] > 0).sum().detach().cpu()
            for j in range(len(lower_bounds))
    ])
    n_total = sum([lower_bounds[j].numel() for j in range(len(lower_bounds))])

    return n_total - n_unstable, n_unstable, lower_bounds, upper_bounds



from beartype import beartype
import torch

from ...onnx2pytorch.convert.model import ConvertModel

# @beartype
@torch.no_grad()
def check_solution(net: ConvertModel, adv: torch.Tensor,
                   cs: torch.Tensor, rhs: torch.Tensor,
                   data_min: torch.Tensor, data_max: torch.Tensor) -> torch.Tensor:
    old_dtype = adv.dtype
    adv = adv.to(data_min.dtype)
    adv = torch.clamp(torch.clamp(adv, max=data_max), min=data_min)
    assert torch.all(data_min <= adv) and torch.all(adv <= data_max)
    net.to(data_min.dtype)
    output = net(adv).detach()
    cond = torch.matmul(cs, output.unsqueeze(-1)).squeeze(-1) - rhs
    net.to(old_dtype)
    # print(cond.detach().cpu().numpy())
    return (cond.amax(dim=-1, keepdim=True) < 0.0).any(dim=-1).any(dim=-1)



from .base import *
from .solver_utils import grb


class BoundAdd(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        options = options or {}
        # FIXME: This is not the right way to enable patches mode.
        # Instead we must traverse the graph and determine when patches mode needs to be used.

        self.mode = options.get("conv_mode", "matrix")

    def forward(self, x, y):
        self.x_shape = x.shape
        self.y_shape = y.shape
        return x + y

    def bound_backward(self, last_lA, last_uA, x, y, **kwargs):
        def _bound_oneside(last_A, w):
            if last_A is None:
                return None
            return self.broadcast_backward(last_A, w)

        uA_x = _bound_oneside(last_uA, x)
        uA_y = _bound_oneside(last_uA, y)
        lA_x = _bound_oneside(last_lA, x)
        lA_y = _bound_oneside(last_lA, y)
        return [(lA_x, uA_x), (lA_y, uA_y)], 0, 0

    def bound_forward(self, dim_in, x, y):
        lb, ub = x.lb + y.lb, x.ub + y.ub

        def add_w(x_w, y_w, x_b, y_b):
            if x_w is None and y_w is None:
                return None
            elif x_w is not None and y_w is not None:
                return x_w + y_w
            elif y_w is None:
                return x_w + torch.zeros_like(y_b)
            else:
                return y_w + torch.zeros_like(x_b)

        lw = add_w(x.lw, y.lw, x.lb, y.lb)
        uw = add_w(x.uw, y.uw, x.ub, y.ub)

        return LinearBound(lw, lb, uw, ub)

    def interval_propagate(self, x, y):
        assert (not isinstance(y, Tensor))
        return x[0] + y[0], x[1] + y[1]

    def build_solver(self, *v, model, C=None, model_type="mip", solver_pkg="gurobi"):
        if isinstance(v[0], Tensor) and isinstance(v[1], Tensor):
            # constants if both inputs are tensors
            self.solver_vars = self.forward(v[0], v[1])
            return
        # we have both gurobi vars as inputs
        this_layer_shape = self.output_shape
        gvar_array1 = np.array(v[0])
        gvar_array2 = np.array(v[1])
        assert gvar_array1.shape == gvar_array2.shape and gvar_array1.shape == this_layer_shape[1:]

        # flatten to create vars and constrs first
        gvar_array1 = gvar_array1.reshape(-1)
        gvar_array2 = gvar_array2.reshape(-1)
        new_layer_gurobi_vars = []
        for neuron_idx, (var1, var2) in enumerate(zip(gvar_array1, gvar_array2)):
            var = model.addVar(lb=-float('inf'), ub=float('inf'), obj=0,
                            vtype=grb.GRB.CONTINUOUS,
                            name=f'lay{self.name}_{neuron_idx}')
            model.addConstr(var == (var1 + var2), name=f'lay{self.name}_{neuron_idx}_eq')
            new_layer_gurobi_vars.append(var)

        # reshape to the correct list shape of solver vars
        self.solver_vars = np.array(new_layer_gurobi_vars).reshape(this_layer_shape[1:]).tolist()
        model.update()


class BoundSub(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        # FIXME: This is not the right way to enable patches mode. Instead we must traverse the graph and determine when patches mode needs to be used.
        self.mode = options.get("conv_mode", "matrix")

    def forward(self, x, y):
        self.x_shape = x.shape
        self.y_shape = y.shape
        return x - y

    def bound_backward(self, last_lA, last_uA, x, y, **kwargs):
        def _bound_oneside(last_A, w, sign=-1):
            if last_A is None:
                return None
            if isinstance(last_A, torch.Tensor):
                return self.broadcast_backward(sign * last_A, w)
            elif isinstance(last_A, Patches):
                if sign == 1:
                    # Patches shape requires no broadcast.
                    return last_A
                else:
                    # Multiply by the sign.
                    return last_A.create_similar(sign * last_A.patches)
            else:
                raise ValueError(f'Unknown last_A type {type(last_A)}')

        uA_x = _bound_oneside(last_uA, x, sign=1)
        uA_y = _bound_oneside(last_uA, y, sign=-1)
        lA_x = _bound_oneside(last_lA, x, sign=1)
        lA_y = _bound_oneside(last_lA, y, sign=-1)
        return [(lA_x, uA_x), (lA_y, uA_y)], 0, 0

    def bound_forward(self, dim_in, x, y):
        lb, ub = x.lb - y.ub, x.ub - y.lb

        def add_w(x_w, y_w, x_b, y_b):
            if x_w is None and y_w is None:
                return None
            elif x_w is not None and y_w is not None:
                return x_w + y_w
            elif y_w is None:
                return x_w + torch.zeros_like(y_b)
            else:
                return y_w + torch.zeros_like(x_b)

        lw = add_w(x.lw, -y.uw, x.lb, y.lb)
        uw = add_w(x.uw, -y.lw, x.ub, y.ub)

        return LinearBound(lw, lb, uw, ub)

    def interval_propagate(self, x, y):
        return x[0] - y[1], x[1] - y[0]

    def build_solver(self, *v, model, C=None, model_type="mip", solver_pkg="gurobi"):
        if isinstance(v[0], Tensor) and isinstance(v[1], Tensor):
            # constants if both inputs are tensors
            self.solver_vars = self.forward(v[0], v[1])
            return
        # we have both gurobi vars as inputs
        this_layer_shape = self.output_shape
        gvar_array1 = np.array(v[0])
        gvar_array2 = np.array(v[1])
        assert gvar_array1.shape == gvar_array2.shape and gvar_array1.shape == this_layer_shape[1:]

        # flatten to create vars and constrs first
        gvar_array1 = gvar_array1.reshape(-1)
        gvar_array2 = gvar_array2.reshape(-1)
        new_layer_gurobi_vars = []
        for neuron_idx, (var1, var2) in enumerate(zip(gvar_array1, gvar_array2)):
            var = model.addVar(lb=-float('inf'), ub=float('inf'), obj=0,
                            vtype=grb.GRB.CONTINUOUS,
                            name=f'lay{self.name}_{neuron_idx}')
            model.addConstr(var == (var1 - var2), name=f'lay{self.name}_{neuron_idx}_eq')
            new_layer_gurobi_vars.append(var)

        # reshape to the correct list shape of solver vars
        self.solver_vars = np.array(new_layer_gurobi_vars).reshape(this_layer_shape[1:]).tolist()
        model.update()



import torch
import warnings
from .bound_ops import *
from .utils import *
from .linear_bound import LinearBound
from .perturbations import PerturbationLpNorm

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .bound_general import BoundedModule

import sys
sys.setrecursionlimit(1000000)


def forward_general(self: 'BoundedModule', C=None, node=None, concretize=False,
                    offset=0):
    if self.bound_opts['dynamic_forward']:
        return self.forward_general_dynamic(C, node, concretize, offset)

    if C is None:
        if hasattr(node, 'linear'):
            return node.linear.lower, node.linear.upper
        if not node.from_input:
            node.linear = LinearBound(None, node.value, None, node.value, node.value, node.value)
            return node.value, node.value
        if not node.perturbed:
            node.lower = node.upper = self.get_forward_value(node)
        if hasattr(node, 'lower'):
            node.linear = LinearBound(None, node.lower, None, node.upper, node.lower, node.upper)
            return node.lower, node.upper

    for l_pre in node.inputs:
        if not hasattr(l_pre, 'linear'):
            self.forward_general(node=l_pre, offset=offset)
    inp = [l_pre.linear for l_pre in node.inputs]
    node._start = '_forward'
    if (C is not None and isinstance(node, BoundLinear) and
            not node.is_input_perturbed(1) and not node.is_input_perturbed(2)):
        linear = node.bound_forward(self.dim_in, *inp, C=C)
        C_merged = True
    else:
        linear = node.linear = node.bound_forward(self.dim_in, *inp)
        C_merged = False

    lw, uw = linear.lw, linear.uw
    lower, upper = linear.lb, linear.ub

    if C is not None and not C_merged:
        # FIXME use bound_forward of BoundLinear
        C_pos, C_neg = C.clamp(min=0), C.clamp(max=0)
        _lw = torch.matmul(lw, C_pos.transpose(-1, -2)) + torch.matmul(uw, C_neg.transpose(-1, -2))
        _uw = torch.matmul(uw, C_pos.transpose(-1, -2)) + torch.matmul(lw, C_neg.transpose(-1, -2))
        lw, uw = _lw, _uw
        _lower = torch.matmul(lower.unsqueeze(1), C_pos.transpose(-1, -2)) + \
                    torch.matmul(upper.unsqueeze(1), C_neg.transpose(-1, -2))
        _upper = torch.matmul(upper.unsqueeze(1), C_pos.transpose(-1, -2)) + \
                    torch.matmul(lower.unsqueeze(1), C_neg.transpose(-1, -2))
        lower, upper = _lower.squeeze(1), _upper.squeeze(1)

    logger.debug(f'Forward bounds to {node}')

    if concretize:
        if lw is not None or uw is not None:
            roots = self.roots()
            prev_dim_in = 0
            batch_size = lw.shape[0]
            assert (lw.ndim > 1)
            lA = lw.reshape(batch_size, self.dim_in, -1).transpose(1, 2)
            uA = uw.reshape(batch_size, self.dim_in, -1).transpose(1, 2)
            for i in range(len(roots)):
                if hasattr(roots[i], 'perturbation') and roots[i].perturbation is not None:
                    _lA = lA[:, :, prev_dim_in : (prev_dim_in + roots[i].dim)]
                    _uA = uA[:, :, prev_dim_in : (prev_dim_in + roots[i].dim)]
                    lower = lower + roots[i].perturbation.concretize(
                        roots[i].center, _lA, sign=-1, aux=roots[i].aux).view(lower.shape)
                    upper = upper + roots[i].perturbation.concretize(
                        roots[i].center, _uA, sign=+1, aux=roots[i].aux).view(upper.shape)
                    prev_dim_in += roots[i].dim
        linear.lower, linear.upper = lower, upper

        if C is None:
            node.linear = linear
            node.lower, node.upper = lower, upper

        if self.bound_opts['forward_refinement']:
            need_refinement = False
            for out in node.output_name:
                out_node = self[out]
                for i in getattr(out_node, 'requires_input_bounds', []):
                    if out_node.inputs[i] == node:
                        need_refinement = True
                        break
            if need_refinement:
                self.forward_refinement(node)
        return lower, upper


def forward_general_dynamic(self: 'BoundedModule', C=None, node=None,
                            concretize=False, offset=0):
    max_dim = self.bound_opts['forward_max_dim']

    if C is None:
        if hasattr(node, 'linear'):
            assert not concretize

            linear = node.linear
            if offset == 0:
                if linear.lw is None:
                    return linear
                elif linear.lw.shape[1] <= max_dim:
                    return linear
            if linear.lw is not None:
                lw = linear.lw[:, offset:offset+max_dim]
                x_L = linear.x_L[:, offset:offset+max_dim]
                x_U = linear.x_U[:, offset:offset+max_dim]
                tot_dim = linear.tot_dim
                if offset == 0:
                    lb = linear.lb
                else:
                    lb = torch.zeros_like(linear.lb)
            else:
                lw = x_L = x_U = None
                tot_dim = 0
                lb = linear.lb
            return LinearBound(
                lw, lb, lw, lb, x_L=x_L, x_U=x_U,
                offset=offset, tot_dim=tot_dim,
            )

        # These cases have no coefficient tensor
        if not node.from_input:
            if concretize:
                return node.value, node.value
            else:
                node.linear = LinearBound(
                    None, node.value, None, node.value, node.value, node.value)
                return node.linear
        if not node.perturbed:
            if not hasattr(node, 'lower'):
                node.lower = node.upper = self.get_forward_value(node)
            if concretize:
                return node.lower, node.upper
            else:
                if offset > 0:
                    lb = torch.zeros_like(node.lower)
                else:
                    lb = node.lower
                node.linear = LinearBound(None, lb, None, lb, node.lower, node.upper)
                return node.linear

    if offset == 0:
        logger.debug(f'forward_general_dynamic: node={node}')

    inp = []
    for l_pre in node.inputs:
        linear_inp = self.forward_general_dynamic(node=l_pre, offset=offset)
        linear_inp.lower = getattr(l_pre, 'lower', None)
        linear_inp.upper = getattr(l_pre, 'upper', None)
        inp.append(linear_inp)
    node._start = '_forward'
    if (C is not None and isinstance(node, BoundLinear) and
            not node.is_input_perturbed(1) and not node.is_input_perturbed(2)):
        linear = node.bound_dynamic_forward(
            *inp, C=C, max_dim=max_dim, offset=offset)
        C_merged = True
    else:
        linear = node.bound_dynamic_forward(
            *inp, max_dim=max_dim, offset=offset)
        C_merged = False
    if offset > 0:
        linear.lb = linear.ub = torch.zeros_like(linear.lb)

    lw, lb, tot_dim = linear.lw, linear.lb, linear.tot_dim
    #logger.debug(f'forward_general_dynamic: node={node}, w_size={lw.shape[1]}, tot_dim={tot_dim}')

    if C is not None and not C_merged:
        # FIXME use bound_forward of BoundLinear
        lw = torch.matmul(lw, C.transpose(-1, -2))
        lb = torch.matmul(lb.unsqueeze(1), C.transpose(-1, -2)).squeeze(1)

    if concretize:
        lower = upper = lb
        if lw is not None:
            batch_size = lw.shape[0]
            assert (lw.ndim > 1)
            if lw.shape[1] > 0:
                A = lw.reshape(batch_size, lw.shape[1], -1).transpose(1, 2)
                ptb = PerturbationLpNorm(x_L=linear.x_L, x_U=linear.x_U)
                lower = lower + ptb.concretize(x=None, A=A, sign=-1).view(lb.shape)
                upper = upper + ptb.concretize(x=None, A=A, sign=1).view(lb.shape)
            offset_next = offset + max_dim
            more = offset_next < tot_dim
        else:
            more = False

        if C is None and offset == 0 and not more:
            node.linear = linear

        if more:
            if lw is not None and lw.shape[1] > 0:
                del A
                del ptb
                del lw
                del linear
            del inp
            # TODO make it non-recursive
            lower_next, upper_next = self.forward_general_dynamic(
                C, node, concretize=True, offset=offset_next)
            lower = lower + lower_next
            upper = upper + upper_next

        if C is None:
            node.lower, node.upper = lower, upper

        return lower, upper
    else:
        return linear


def clean_memory(self: 'BoundedModule', node):
    """ Remove linear bounds that are no longer needed. """
    # TODO add an option to retain these bounds
    for inp in node.inputs:
        if hasattr(inp, 'linear') and inp.linear is not None:
            clean = True
            for out in inp.output_name:
                out_node = self[out]
                if not (hasattr(out_node, 'linear') and out_node.linear is not None):
                    clean = False
            if clean:
                if isinstance(inp.linear, tuple):
                    for item in inp.linear:
                        del item
                delattr(inp, 'linear')


def forward_refinement(self: 'BoundedModule', node):
    """ Refine forward bounds with backward bound propagation
    (only refine unstable positions). """
    unstable_size_before = torch.logical_and(node.lower < 0, node.upper > 0).sum()
    if unstable_size_before == 0:
        return
    unstable_idx, unstable_size = self.get_unstable_locations(
        node.lower, node.upper, conv=isinstance(node, BoundConv))
    logger.debug(f'Forward refinement for {node}')
    batch_size = node.lower.shape[0]
    ret = self.batched_backward(
        node, C=None, unstable_idx=unstable_idx, batch_size=batch_size)
    self.restore_sparse_bounds(
        node, unstable_idx, unstable_size, node.lower, node.upper,
        new_lower=ret[0], new_upper=ret[1])
    unstable_size_after = torch.logical_and(node.lower < 0, node.upper > 0).sum()
    logger.debug(f'  Unstable neurons: {unstable_size_before} -> {unstable_size_after}')
    # TODO also update linear bounds?


def init_forward(self: 'BoundedModule', roots, dim_in):
    if dim_in == 0:
        raise ValueError("At least one node should have a specified perturbation")
    prev_dim_in = 0
    # Assumption: roots[0] is the input node which implies batch_size
    batch_size = roots[0].value.shape[0]
    dynamic = self.bound_opts['dynamic_forward']
    for i in range(len(roots)):
        if hasattr(roots[i], 'perturbation') and roots[i].perturbation is not None:
            shape = roots[i].linear.lw.shape
            if dynamic:
                if shape[1] != dim_in:
                    raise NotImplementedError('Dynamic forward bound is not supported yet when there are multiple perturbed inputs.')
                ptb = roots[i].perturbation
                if (type(ptb) != PerturbationLpNorm or ptb.norm < np.inf
                        or ptb.x_L is None or ptb.x_U is None):
                    raise NotImplementedError(
                        'For dynamic forward bounds, only Linf (box) perturbations are supported, and x_L and x_U must be explicitly provided.')
                roots[i].linear.x_L = (
                    ptb.x_L_sparse.view(batch_size, -1) if ptb.sparse
                    else ptb.x_L.view(batch_size, -1))
                roots[i].linear.x_U = (
                    ptb.x_U_sparse.view(batch_size, -1) if ptb.sparse
                    else ptb.x_U.view(batch_size, -1))
            else:
                lw = torch.zeros(shape[0], dim_in, *shape[2:]).to(roots[i].linear.lw)
                lw[:, prev_dim_in:(prev_dim_in+shape[1])] = roots[i].linear.lw
                if roots[i].linear.lw.data_ptr() == roots[i].linear.uw.data_ptr():
                    uw = lw
                else:
                    uw = torch.zeros(shape[0], dim_in, *shape[2:]).to(roots[i].linear.uw)
                    uw[:, prev_dim_in:(prev_dim_in+shape[1])] = roots[i].linear.uw
                roots[i].linear.lw = lw
                roots[i].linear.uw = uw
            if i >= self.num_global_inputs:
                roots[i].forward_value = roots[i].forward_value.unsqueeze(0).repeat(
                    *([batch_size] + [1] * self.forward_value.ndim))
            prev_dim_in += shape[1]
        else:
            # FIXME: double check
            if roots[i].forward_value.ndim == 0:
                roots[i].forward_value = roots[i].forward_value.unsqueeze(0)
            
            b = fv = roots[i].forward_value
            shape = fv.shape
            if roots[i].from_input:
                w = torch.zeros(shape[0], dim_in, *shape[1:], device=self.device)
                # warnings.warn(f'Creating a LinearBound with zero weights with shape {w.shape}')
            else:
                w = None
            roots[i].linear = LinearBound(w, b, w, b, b, b)
            roots[i].lower = roots[i].upper = b



"""RNN."""
from .base import *


class BoundRNN(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.complex = True
        self.output_index = output_index
        raise NotImplementedError(
            'torch.nn.RNN is not supported at this time.'
            'Please implement your RNN with torch.nn.RNNCell and a manual for-loop.'
            'See an example of LSTM:'
            'https://github.com/Verified-Intelligence/auto_LiRPA/blob/10a9b30/examples/sequence/lstm.py#L9')

    def forward(self, x, weight_input, weight_recurrent, bias, sequence_length, initial_h):
        assert (torch.sum(torch.abs(initial_h)) == 0)

        self.input_size = x.shape[-1]
        self.hidden_size = weight_input.shape[-2]

        class BoundRNNImpl(nn.Module):
            def __init__(self, input_size, hidden_size,
                         weight_input, weight_recurrent, bias, output_index):
                super().__init__()

                self.input_size = input_size
                self.hidden_size = hidden_size

                self.cell = torch.nn.RNNCell(
                    input_size=input_size,
                    hidden_size=hidden_size
                )

                self.cell.weight_ih.data.copy_(weight_input.squeeze(0).data)
                self.cell.weight_hh.data.copy_(weight_recurrent.squeeze(0).data)
                self.cell.bias_ih.data.copy_((bias.squeeze(0))[:hidden_size].data)
                self.cell.bias_hh.data.copy_((bias.squeeze(0))[hidden_size:].data)

                self.output_index = output_index

            def forward(self, x, hidden):
                length = x.shape[0]
                outputs = []
                for i in range(length):
                    hidden = self.cell(x[i, :], hidden)
                    outputs.append(hidden.unsqueeze(0))
                outputs = torch.cat(outputs, dim=0)

                if self.output_index == 0:
                    return outputs
                else:
                    return hidden

        self.model = BoundRNNImpl(
            self.input_size, self.hidden_size,
            weight_input, weight_recurrent, bias,
            self.output_index)
        self.input = (x, initial_h)

        return self.model(*self.input)


from beartype import beartype
import torch

# @beartype
def serialize_specs(x: torch.Tensor, cs: torch.Tensor, rhs: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor, list[list[int]]]:
    assert x.shape[0] == 1, print(f'Invalid batch size: {x.shape[0]}')
    cs_mat, rhs_mat, cond_mat = [], [], []
    for cs_i, rhs_i in zip(cs, rhs):
        cs_mat.append(torch.Tensor(cs_i).to(x.device))
        rhs_mat.append(torch.Tensor(rhs_i).to(x.device))
        cond_mat.append(rhs_i.shape[0]) # mark the `and` group
    cs_mat = torch.cat(cs_mat, dim=0).unsqueeze(0).to(x.device)
    rhs_mat = torch.cat(rhs_mat, dim=0).unsqueeze(0).to(x.device)
    return cs_mat, rhs_mat, [cond_mat]


# @beartype
def check_adv_multi(input: torch.Tensor, output: torch.Tensor,
                    serialized_conditions: tuple[torch.Tensor, torch.Tensor, list[list[int]]],
                    data_max: torch.Tensor, data_min: torch.Tensor) -> torch.Tensor:
    cs_mat, rhs_mat, cond_mat = serialized_conditions
    if len(cond_mat[0]) == rhs_mat.shape[-1]:
        assert all([cond_mat[0][0] == i for i in cond_mat[0]])
        cs_mat = cs_mat.view(cs_mat.shape[0], 1, len(cond_mat[0]), -1, cs_mat.shape[-1])
        # [batch_size, restarts, num_or_spec, num_and_spec, output_dim]
        rhs_mat = rhs_mat.view(rhs_mat.shape[0], 1, len(cond_mat[0]), -1)
        cond = torch.matmul(cs_mat, output.unsqueeze(-1)).squeeze(-1) - rhs_mat

        valid = ((input <= data_max) & (input >= data_min))
        valid = valid.view(*valid.shape[:3], -1)
        # [num_example, restarts, num_all_spec, output_dim]
        valid = valid.all(-1).view(valid.shape[0], valid.shape[1], len(cond_mat[0]), -1)
        # [num_example, restarts, num_or_spec, num_and_spec]
        res = ((cond.amax(dim=-1, keepdim=True) < 0.0) & valid).any(dim=-1).any(dim=-1).any(dim=-1)
    else:
        output = output.repeat_interleave(torch.tensor(cond_mat[0]).to(output.device), dim=2)
        # [num_example, num_restarts, num_spec, num_output]

        cs_mat = cs_mat.view(cs_mat.shape[0], 1, -1, cs_mat.shape[-1])
        # [num_example, 1, num_spec, num_output]
        rhs_mat = rhs_mat.view(rhs_mat.shape[0], 1, -1)
        # [num_example, 1, num_spec]

        cond = torch.clamp((cs_mat * output).sum(-1) - rhs_mat, min=0.0)
        # [num_example, 1, num_spec]

        group_C = torch.zeros(len(cond_mat[0]), cs_mat.shape[2], device=cond.device, dtype=cond.dtype) # [num_or_spec, num_total_spec]
        x_index = []
        y_index = []
        index = 0

        for i, num_cond in enumerate(cond_mat[0]):
            x_index.extend([i] * num_cond)
            y_index.extend([index+j] for j in range(num_cond))
            index += num_cond

        group_C[x_index, y_index] = 1.0

        # loss shape: [batch_size, num_restarts, num_total_spec]
        cond = group_C.matmul(cond.unsqueeze(-1)).squeeze(-1)

        valid = ((input <= data_max) & (input >= data_min))
        valid = valid.view(*valid.shape[:3], -1)
        # [num_example, restarts, num_all_spec, output_dim]
        valid = valid.all(-1).view(valid.shape[0], valid.shape[1], len(cond_mat[0]), -1)
        # [num_example, restarts, num_or_spec, num_and_spec]

        valid = valid.all(-1)

        # [num_example, num_restarts, num_or_example]
        res = ((cond == 0.0) & valid).any(dim=-1).any(dim=-1)

    return res.all()


# @beartype
def get_loss(origin_out: torch.Tensor | None, output: torch.Tensor,
             serialized_conditions: tuple[torch.Tensor, torch.Tensor, list[list[int]]],
             gama_lambda: float = 0.0, threshold: float = -1e-5) -> torch.Tensor:
    cs_mat, rhs_mat, cond_mat = serialized_conditions
    if rhs_mat.shape[-1] == len(cond_mat[0]):
        assert all([cond_mat[0][0] == i for i in cond_mat[0]])
        cs_mat = cs_mat.view(cs_mat.shape[0], 1, output.shape[2], -1, cs_mat.shape[-1])
        # [num_example, 1, num_or_spec, num_and_spec, num_output]
        rhs_mat = rhs_mat.view(rhs_mat.shape[0], 1, output.shape[2], -1)
        loss = cs_mat.matmul(output.unsqueeze(-1)).squeeze(-1) - rhs_mat
        loss = torch.clamp(loss, min=threshold)
        # [num_example, num_restarts, num_or_spec, num_and_spec]
        loss = -loss
    else:
        output = output.repeat_interleave(torch.tensor(cond_mat[0]).to(output.device), dim=2)
        if origin_out is not None:
            origin_out = origin_out.repeat_interleave(torch.tensor(cond_mat[0]).to(output.device), dim=2)
        # [num_example, num_restarts, num_spec, num_output]

        cs_mat = cs_mat.view(cs_mat.shape[0], 1, -1, cs_mat.shape[-1])
        # [num_example, 1, num_spec, num_output]
        rhs_mat = rhs_mat.view(rhs_mat.shape[0], 1, -1)
        # [num_example, 1, num_spec]

        loss = (cs_mat * output).sum(-1) - rhs_mat
        loss = torch.clamp(loss, min=threshold)
        loss = -loss

    if origin_out is not None:
        loss_gamma = loss.sum() + (gama_lambda * (output - origin_out)**2).sum(dim=3).sum()
    else:
        loss_gamma = loss.sum()
    # [num_example, num_restarts, num_or_spec, num_and_spec]
    return loss_gamma



import torch.nn.functional as F

from ..operations.base import Operator


class Pad(Operator):
    def __init__(self, mode="constant", padding=None, constant=0):
        self.mode = mode
        self.padding = padding
        super().__init__()

    def forward(self, input, pads=None, value=0):
        if self.padding is not None:
            pads = self.padding
        elif pads is None:
            raise TypeError("forward() missing 1 required positional argument: 'pads'")
        assert input.ndim == 4
        if len(pads) > 4:
            # pads should be [0,0,pad_top,pad_left,0,0,pad_bottom,pad_right]
            assert pads[0] == pads[1] == pads[4] == pads[5] == 0
            pads = [int(pads[3]), int(pads[7]), int(pads[2]), int(pads[6])]
        out = F.pad(input, list(pads), mode=self.mode, value=value)
        return out

    def extra_repr(self) -> str:
        return "mode={}, padding={}".format(self.mode, self.padding)



from collections import namedtuple


AbstractResults = namedtuple(
    'AbstractResults', 
    (
        'objective_ids',
        'output_lbs', 
        'masks', 'lAs', 'histories', 
        'lower_bounds', 'upper_bounds', 
        'input_lowers', 'input_uppers',
        'slopes', 'betas', 
        'cs', 'rhs',
        'sat_solvers'
    ), 
    defaults=(None,) * 14
)


class ReturnStatus:

    UNSAT   = 'unsat'
    SAT     = 'sat'
    UNKNOWN = 'unknown'
    TIMEOUT = 'timeout'
    RESTART = 'restart'



from sortedcontainers import SortedList
import gurobipy as grb
import numpy as np
import random
import torch
import time
import math
import copy

from ..util.misc.result import AbstractResults
from ..util.misc.check import check_solution
from ..util.misc.logger import logger
from ..setting import Settings

multiprocess_mip_attack_model = None
multiprocess_stop = False


class CounterExample:

    "A single example"

    def __init__(self, x, obj, pattern):
        self.x = x
        self.obj = obj  # smaller obj is better
        self.activation_pattern = pattern

    def __lt__(self, other):
        return self.obj < other.obj

    def __le__(self, other):
        return self.obj <= other.obj

    def __eq__(self, other):
        return self.obj == other.obj


class CounterExamplePool:

    def __init__(self, abstractor, objectives, capacity):
        self.net = abstractor.pytorch_model
        self.objectives = objectives
        self.input_shape = abstractor.input_shape
        self.device = abstractor.device

        self.cs = objectives.cs.transpose(0, 1).cpu().clone() # [1, #props, #outputs]

        # maximum number of cex in pool
        self.pool = SortedList()
        self.capacity = capacity
        self.abstractor = abstractor

        # init pool
        self.threshold = None
        self._initialize()


    def _initialize(self):
        input_lowers = self.objectives.lower_bounds.view(-1, *self.input_shape[1:]).to(self.device)
        input_uppers = self.objectives.upper_bounds.view(-1, *self.input_shape[1:]).to(self.device)

        repeat = self.capacity // len(input_lowers) + 1 if (self.capacity % len(input_lowers)) else self.capacity // len(input_lowers)
        repeat *= 3

        input_lowers_repeat = input_lowers.repeat(repeat, *[1]*(len(self.input_shape) - 1))
        input_uppers_repeat = input_uppers.repeat(repeat, *[1]*(len(self.input_shape) - 1))

        # seed = random.randint(1, 100000)
        # seed = 28998
        # print('seed:', seed)
        # torch.manual_seed(seed)
        adv_example = (input_uppers_repeat - input_lowers_repeat) * torch.rand(input_lowers_repeat.shape, device=self.device) + input_lowers_repeat
        # adv_example = torch.clamp(torch.clamp(torch.randn(input_lowers_repeat.shape, device=self.device), max=input_uppers_repeat), min=input_lowers_repeat)
        assert (adv_example >= input_lowers_repeat).all() and (adv_example <= input_uppers_repeat).all()
        # print(adv_example.sum().item())
        self.add(adv_example)
        # exit()

    @torch.no_grad()
    def add(self, inputs):
        assert len(inputs) > 0
        # reset_perturbed_nodes must be False otherwise the .perturbed property will be missing.
        pred = self.abstractor.net(inputs).cpu()
        # FIXME: add rhs
        pred = pred.matmul(self.cs[0].transpose(-1, -2)).min(-1).values

        activations = [None] * len(self.abstractor.net.relus)
        for layer_i, layer in enumerate(self.abstractor.net.relus):
            activations[layer_i] = (layer.inputs[0].forward_value.flatten(1) > 0).int().cpu()

        for batch_idx in range(len(inputs)):
            activation_i = [_[batch_idx] for _ in activations]
            self._add_one(CounterExample(inputs[batch_idx], pred[batch_idx].item(), activation_i))


    def _add_one(self, cex):
        if len(self.pool) >= self.capacity:
            if cex.obj >= self.pool[-1].obj:
                # skip if cex is worse than the worst example in pool
                print('skip', cex.obj, self.pool[-1].obj)
                return

            pop_idx = -1
            if 1:
                current_patterns = torch.stack([torch.cat([ii.flatten() for ii in i.activation_pattern]) for i in self.pool])
                this_pattern = torch.cat([ii.flatten() for ii in cex.activation_pattern]).view(1, -1)
                # print(current_patterns)
                # print(this_pattern)

                diff = torch.cdist(current_patterns.float(), this_pattern.float(), p=0).flatten()
                min_idx = diff.argmin()

                if self.threshold is None:
                    # we set the threshold as the lowest diff when the first time we need to filter
                    self.threshold = diff[min_idx] if diff[min_idx] > 1.0 else 1.0

                if diff[min_idx] < self.threshold and cex.obj < self.pool[min_idx].obj:
                    pop_idx = min_idx.item()

            # print('remove idx:', pop_idx)
            # assert pop_idx == -1
            self.pool.pop(pop_idx)

        self.pool.add(cex)


    def get_common_pattern(self, prob_threshold=0.5):
        selected_advs = self.pool
        pos_threshold = min(int(math.ceil(prob_threshold * len(selected_advs))), len(selected_advs))
        neg_threshold = len(selected_advs) - pos_threshold

        # collect all activation patterns
        all_patterns = [[] for i in range(len(self.abstractor.net.relus))]
        for adv in selected_advs:
            for layer_i in range(len(all_patterns)):
                all_patterns[layer_i].append(adv.activation_pattern[layer_i])
        # print(all_patterns)

        # concat activation patterns across examples
        ret = [None] * len(self.abstractor.net.relus)
        for layer_i in range(len(all_patterns)):
            layer_i_pattern = torch.stack(all_patterns[layer_i], dim=0)
            # print(layer_i, layer_i_pattern.shape)
            acc_pattern = layer_i_pattern.sum(dim=0)
            # print(layer_i, acc_pattern)
            # +1 for most neurons active, -1 for most neurons inactive, 0 for not selected
            ret[layer_i] = ((acc_pattern >= pos_threshold).int() - (acc_pattern < neg_threshold).int())

        return ret

    def get_uncommon_pattern(self, prob_threshold=0.5):
        pass

    def __len__(self):
        return len(self.pool)

def mip_solver_worker(candidate, n_inputs):
    global multiprocess_stop
    if multiprocess_stop:
        return None

    tmp_model = multiprocess_mip_attack_model.copy()

    v = tmp_model.getVarByName(candidate)
    vlb = out_lb = v.lb
    vub = out_ub = v.ub
    adv = None

    if vlb > 0:
        return None

    tmp_model.setObjective(v, grb.GRB.MINIMIZE)
    tmp_model.update()
    try:
        tmp_model.optimize()
    except grb.GurobiError as e:
        print(f'Gurobi error: {e.message}')
        return None
    except KeyboardInterrupt:
        multiprocess_stop = True
        return None


    vlb = max(tmp_model.objbound, out_lb)
    if tmp_model.solcount > 0:
        vub = min(tmp_model.objval, out_ub)
    if vub < 0:
        input_vars = [tmp_model.getVarByName(f'inp_{dim}') for dim in range(n_inputs)]
        adv = [var.X for var in input_vars]
        multiprocess_stop = True


    print(f'[!] Attacking {candidate} [{vlb, vub}], status: {tmp_model.status}, #vars: {tmp_model.NumVars}, #constrs: {tmp_model.NumConstrs}')

    if tmp_model.status in [3, 11]: # infeasible
        multiprocess_stop = True

    return adv


class MIPAttacker:

    def __init__(self, abstractor, objectives):
        self.net = abstractor.pytorch_model
        self.input_shape = abstractor.input_shape
        self.device = abstractor.device

        assert objectives.cs.shape[1] == 1 # c shape: [#props, 1, #outputs]
        self.objectives = copy.deepcopy(objectives)
        self.objectives.lower_bounds = self.objectives.lower_bounds.view(-1, *self.input_shape[1:]).to(self.device)
        self.objectives.upper_bounds = self.objectives.upper_bounds.view(-1, *self.input_shape[1:]).to(self.device)
        self.objectives.cs = self.objectives.cs.to(self.device)
        self.objectives.rhs = self.objectives.rhs.to(self.device)

        self.abstractor = abstractor
        self.pre_relu_names = {i: layer.inputs[0].name for (i, layer) in enumerate(self.abstractor.net.perturbed_optimizable_activations)}
        self.relu_names = {i: layer.name for (i, layer) in enumerate(self.abstractor.net.perturbed_optimizable_activations)}

        # if (not hasattr(self.abstractor, 'model')) or ('mip' not in self.abstractor.net.model.ModelName):
        #     self.init_mip_model(
        #         input_lowers=self.objectives.lower_bounds[0:1],
        #         input_uppers=self.objectives.upper_bounds[0:1],
        #         cs=self.objectives.cs.transpose(0, 1),
        #     )

        # self.mip_model = self.abstractor.net.model.copy()
        # self.mip_model.setParam('BestBdStop', 1e-5)  # Terminiate as long as we find a positive lower bound.
        # self.mip_model.setParam('BestObjStop', -1e-5)  # Terminiate as long as we find a adversarial example.
        # self.mip_model.setParam('TimeLimit', 5.0)
        # # self.mip_model.setParam('Threads', 1)
        # self.mip_model.update()

        # self.output_names = [v.VarName for v in self.abstractor.net[self.abstractor.net.final_name].solver_vars][-1:]
        # print(self.output_names)
        # exit()


    def init_mip_model(self, input_lowers, input_uppers, cs):
        # print(input_lowers.shape)
        # print(cs.shape)

        tic = time.time()
        self.abstractor.build_lp_solver(
            model_type='mip',
            input_lower=input_lowers,
            input_upper=input_uppers,
            c=cs,
            refine=False,
            timeout=None,
        )
        print(f'Initialize new MIP model in {time.time() - tic} seconds')



    def manual_seed(self, seed):
        self.seed = seed
        random.seed(self.seed)
        torch.manual_seed(self.seed)


    def run(self, reference_bounds=None, timeout=5.0):
        return False, None
        global multiprocess_mip_attack_model, multiprocess_stop
        multiprocess_mip_attack_model = self.mip_model.copy()
        multiprocess_stop = False

        for var_name in self.output_names:
            tic = time.time()
            adv = mip_solver_worker(var_name, np.prod(self.input_shape))
            print('attacking:', var_name, time.time() - tic)
            if adv is not None:
                adv = torch.tensor(adv, device=self.device).view(self.input_shape)
                if check_solution(net=self.net, adv=adv, cs=self.objectives.cs, rhs=self.objectives.rhs, data_min=self.objectives.lower_bounds, data_max=self.objectives.upper_bounds):
                    return True, adv

        return False, None


    def attack_domains(self, domain_params, concretize_percent=0.7):
        batch = len(domain_params.lower_bounds[0])
        logger.debug(f'MIP Attacking: {batch}')

        if batch == 0:
            return

        # worst bounds
        unified_lower_bounds = [_.min(dim=0).values.flatten() for _ in domain_params.lower_bounds[:-1]]
        unified_upper_bounds = [_.max(dim=0).values.flatten() for _ in domain_params.upper_bounds[:-1]]

        unified_bound_shapes = [_.size() for _ in domain_params.lower_bounds[:-1]]


        assert all([(u_lb <= o_lb.flatten(1)).all() for u_lb, o_lb in zip(unified_lower_bounds, domain_params.lower_bounds[:-1])])
        assert all([(u_ub >= o_ub.flatten(1)).all() for u_ub, o_ub in zip(unified_upper_bounds, domain_params.upper_bounds[:-1])])

        print([_.shape for _ in unified_lower_bounds])

        unified_masks = [torch.where(lb_ * ub_ < 0)[0].numpy() for (lb_, ub_) in zip(unified_lower_bounds, unified_upper_bounds)]
        unified_indices = [(l_id, n_id) for l_id in range(len(unified_masks)) for n_id in unified_masks[l_id]]

        tic_start = time.time()
        cac = 0
        while cac < 100:
            concrete_indices = random.sample(unified_indices, int(concretize_percent * len(unified_indices)))

            unified_lower_bounds_cl = [_.clone() for _ in unified_lower_bounds]
            unified_upper_bounds_cl = [_.clone() for _ in unified_upper_bounds]

            for l_id, n_id in concrete_indices:
                assert (unified_lower_bounds_cl[l_id][n_id] < 0) and (unified_upper_bounds_cl[l_id][n_id] > 0)
                if unified_lower_bounds_cl[l_id][n_id] + unified_upper_bounds_cl[l_id][n_id] > 0: # toward active
                    unified_lower_bounds_cl[l_id][n_id] = 0.0
                else:
                    unified_upper_bounds_cl[l_id][n_id] = 0.0
            current_model = self.rebuild_mip_model(unified_lower_bounds_cl, unified_upper_bounds_cl, unified_bound_shapes)
            if current_model is None:
                continue

            global multiprocess_mip_attack_model, multiprocess_stop
            multiprocess_mip_attack_model = current_model.copy()
            multiprocess_stop = False


            # for l_id, n_id in concrete_indices:
            #     assert (unified_lower_bounds[l_id][n_id] < 0) and (unified_upper_bounds[l_id][n_id] > 0)
            #     var = multiprocess_mip_attack_model.getVarByName(f"lay{self.pre_relu_names[l_id]}_{n_id}")
            #     a_var = multiprocess_mip_attack_model.getVarByName(f"aReLU{self.relu_names[l_id]}_{n_id}")
            #     assert (var is not None) and (a_var is not None)
            #     if unified_lower_bounds[l_id][n_id] + unified_upper_bounds[l_id][n_id] > 0: # toward active
            #         var.LB = 0.0
            #         # a_var.LB = 1
            #         # a_var.UB = 1
            #     else: # toward inactive
            #         var.UB = 0.0
            #         # a_var.LB = 0
            #         # a_var.UB = 0

            # multiprocess_mip_attack_model.update()

            # print(concrete_indices)

            for var_name in self.output_names:
                tic = time.time()
                adv = mip_solver_worker(var_name, np.prod(self.input_shape))
                print('Attacking:', var_name, time.time() - tic)

            multiprocess_mip_attack_model = None
            current_model = None
            if adv is not None:
                print('\n\n[!] attacked\n\n')
                exit()
            cac += 1



        print('attacked failed:', time.time() - tic_start)


        exit()

        return None
        # exit()
        # print('mip attack', self.run()[0])
        print('attack_domains', len(domain_params.lower_bounds[0]))
        # print( self.objectives.cs)
        # print( self.objectives.cs.shape)
        if not hasattr(self, 'cex_pool'):
            self.cex_pool = CounterExamplePool(self.abstractor, self.objectives, capacity=200)

        select_domain_params = self.select_domains(domain_params, 100)
        print(f'Selected {len(select_domain_params.lower_bounds[0])} domains from {len(domain_params.lower_bounds[0])} domains for MIP hidden attack')
        # for h in domain_params.histories:
        #     print(h)



    def select_domains(self, domain_params, n_candidates=1):
        # print(len(self.cex_pool))
        # print(common_activation)
        # print([_.shape for _ in common_activation])

        # domain_activations = [
        #     ((domain_params.lower_bounds[j] > 0).int() - (domain_params.upper_bounds[j] < 0).int()).flatten(1).cpu()
        #         for j in range(len(domain_params.lower_bounds) - 1)
        # ]
        return
        if n_candidates >= len(domain_params.lower_bounds[0]):
            return domain_params

        common_activation = self.cex_pool.get_common_pattern()
        domain_masks = compute_masks(domain_params.lower_bounds, domain_params.upper_bounds, 'cpu')
        domain_activations = [
            (((domain_params.lower_bounds[j] + domain_params.upper_bounds[j]) > 0).int() - ((domain_params.lower_bounds[j] + domain_params.upper_bounds[j]) <= 0).int()).flatten(1).cpu()
                for j in range(len(domain_masks))
        ]

        domain_scores = torch.zeros([len(domain_masks[0])])
        for ca, da, dm in zip(common_activation, domain_activations, domain_masks):
            domain_scores += (((ca == da) * dm).sum(dim=1) / dm.sum(dim=1))
        domain_scores = domain_scores / len(domain_masks)
        assert (domain_scores <= 1.0).all()

        select_ids = torch.where(domain_scores >= 0.985)[0]
        # print(select_ids)
        if len(select_ids) >= n_candidates:
            select_ids = torch.topk(domain_scores, n_candidates).indices
        else:
            remain_candidates = n_candidates - len(select_ids)
            output_lbs = domain_params.lower_bounds[-1].flatten().cpu()
            normalized_output_lbs = -output_lbs / output_lbs.neg().max()
            # print(normalized_output_lbs, remain_candidates)
            probs = torch.nn.functional.softmax(normalized_output_lbs / 0.1, dim=0)
            extra_ids = probs.multinomial(remain_candidates, replacement=False)
            # print('extra_ids', extra_ids)
            # print('select_ids', select_ids)
            # print('select_ids', )
            select_ids = torch.concat([select_ids, extra_ids]).unique()


            # assert len(extra_ids) + len(select_ids) == n_candidates
        # print('select_ids', select_ids)

        return AbstractResults(**{
            'lower_bounds': [lb[select_ids] for lb in domain_params.lower_bounds],
            'upper_bounds': [ub[select_ids] for ub in domain_params.upper_bounds],
        })




    def rebuild_mip_model(self, refined_lower_bounds, refined_upper_bounds, shapes):
        print('rebuild model')
        intermediate_layer_bounds = {}
        assert len(shapes) == len(refined_lower_bounds) == len(refined_upper_bounds)

        for idx, (l_id, l_name) in enumerate(self.pre_relu_names.items()):

            intermediate_layer_bounds[l_name] = [
                refined_lower_bounds[l_id].to(self.abstractor.device).view(1, *shapes[idx][1:]),
                refined_upper_bounds[l_id].to(self.abstractor.device).view(1, *shapes[idx][1:])
            ]
        # for name, (lbs, ubs) in intermediate_layer_bounds.items():
        #     print(name, lbs.shape)
        # exit()

        self.abstractor.build_lp_solver(
            model_type='mip',
            input_lower=self.objectives.lower_bounds[0:1],
            input_upper=self.objectives.upper_bounds[0:1],
            c=self.objectives.cs.transpose(0, 1),
            rhs=self.objectives.rhs,
            refine=False,
            intermediate_layer_bounds=intermediate_layer_bounds,
        )

        if not hasattr(self.abstractor.net[self.abstractor.net.final_name], 'solver_vars'):
            return None

        self.output_names = [v.VarName for v in self.abstractor.net[self.abstractor.net.final_name].solver_vars]#[-1:]

        current_model = self.abstractor.net.model.copy()
        # current_model.setParam('Threads', 1)
        current_model.setParam('MIPGap', 0.01)
        current_model.setParam('MIPGapAbs', 0.01)
        current_model.setParam('BestBdStop', 1e-5)
        current_model.setParam('BestObjStop', -1e-5)
        current_model.setParam('TimeLimit', 5.0)
        current_model.update()

        return current_model
